"""_polyline.py: Resqml polyline module."""

import logging

log = logging.getLogger(__name__)

import math as maths
import numpy as np

import resqpy.olio.intersection as meet
import resqpy.lines
import resqpy.olio.point_inclusion as pip
import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vu
import resqpy.olio.write_hdf5 as rwh5
import resqpy.olio.xml_et as rqet
import resqpy.lines._common as rql_c
from resqpy.olio.xml_namespaces import curly_namespace as ns


class Polyline(rql_c._BasePolyline):
    """Class for RESQML polyline representation."""

    resqml_type = 'PolylineRepresentation'

    def __init__(self,
                 parent_model,
                 uuid = None,
                 set_bool = None,
                 set_coord = None,
                 set_crs = None,
                 title = None,
                 rep_int_root = None,
                 originator = None,
                 extra_metadata = None):
        """Initialises a new PolylineRepresentation object.

        arguments:
            parent_model (model.Model object): the model which the new PolylineRepresentation belongs to
            uuid (uuid.UUID, optional): the uuid of an existing RESQML PolylineRepresentation from which
                to initialise the resqpy Polyline
            set_bool (boolean, optional): if True, a new polyline created from coordinates is flagged as
                a closed polyline (polygon); ignored if uuid is not None
            set_coord (numpy array of shape (..., 3), optional): an ordered set of xyz values used to define
                a new polyline; ignored if uuid is not None
            set_crs (uuid.UUID, optional): the uuid of a crs to be used when initialising from coordinates;
                ignored if uuid is not None
            title (str, optional): the citation title to use for a new polyline;
                ignored if uuid is not None
            rep_int_root
            originator (str, optional): the name of the person creating the polyline, defaults to login id;
                ignored if uuid is not None
            extra_metadata (dict, optional): string key, value pairs to add as extra metadata for the polyline;
                ignored if uuid is not None

        returns:
            the newly instantiated Polyline object

        :meta common:
        """

        self.model = parent_model
        self.isclosed = set_bool
        self.nodepatch = None
        self.crs_uuid = set_crs
        self.coordinates = None
        self.centre = None
        self.rep_int_root = rep_int_root  # Optional represented interpretation xml root node
        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)

        if self.root is None and all(i is not None for i in [set_bool, set_coord, set_crs, title]):
            # Using data from a polyline set
            assert set_coord.ndim > 1 and 2 <= set_coord.shape[-1] <= 3
            # allow for x,y or x,y,z incoming coordinates but use x,y,z internally
            coord_shape = list(set_coord.shape)
            coord_shape[-1] = 3
            self.coordinates = np.zeros(tuple(coord_shape))
            self.coordinates[..., :set_coord.shape[-1]] = set_coord
            if set_coord.ndim > 2:
                self.coordinates = self.coordinates.reshape((-1, 3))
            assert len(self.coordinates) > 1, 'at least 2 coordinates needed for polyline'
            self.nodepatch = (0, len(self.coordinates))
            assert not any(map(lambda x: x is None, self.nodepatch))  # Required fields - assert neither are None

        # TODO: Add SeismicCoordinates later - optional field
        # TODO: Add LineRole later - optional field

    def _load_from_xml(self):

        assert self.root is not None  # polyline xml node is specified
        poly_root = self.root

        self.title = rqet.citation_title_for_node(poly_root)

        self.extra_metadata = rqet.load_metadata_from_xml(poly_root)

        self.isclosed = rqet.find_tag_bool(poly_root, 'IsClosed')
        assert self.isclosed is not None  # Required field

        patch_node = rqet.find_tag(poly_root, 'NodePatch')
        assert patch_node is not None  # Required field

        geometry_node = rqet.find_tag(patch_node, 'Geometry')
        assert geometry_node is not None  # Required field

        self.crs_uuid = bu.uuid_from_string(rqet.find_nested_tags_text(geometry_node, ['LocalCrs', 'UUID']))
        assert self.crs_uuid is not None  # Required field

        points_node = rqet.find_tag(geometry_node, 'Points')
        assert points_node is not None  # Required field
        rql_c.load_hdf5_array(self, points_node, 'coordinates', tag = 'Coordinates')

        self.nodepatch = (rqet.find_tag_int(patch_node, 'PatchIndex'), rqet.find_tag_int(patch_node, 'Count'))
        assert not any(map(lambda x: x is None, self.nodepatch))  # Required fields - assert neither are None

        self.rep_int_root = self.model.referenced_node(rqet.find_tag(poly_root, 'RepresentedInterpretation'))

    @property
    def rep_int_uuid(self):
        """Returns the uuid of the represented interpretation."""

        # TODO: Track uuid only, not root
        return rqet.uuid_for_part_root(self.rep_int_root)

    @classmethod
    def from_scaled_polyline(cls, original, scaling, title = None, originator = None, extra_metadata = None):
        """Returns a scaled version of the original polyline.

        arguments:
           original (Polyline): the polyline from which the new polyline will be spawned
           scaling (float): the factor by which the original will be scaled
           title (str, optional): the citation title for the new polyline; inherited from
              original if None
           originator (str, optional): the name of the person creating the polyline; inherited
              from original if None
           extra_metadata (dict, optional): extra metadata for the new polyline; inherited from
              original if None

        returns:
           a new Polyline

        notes:
           the scaling factor is applied to vectors radiating from the balanced centre of the
           original polyline to its coordinates; a scaling of 1.0 will result in a copy of the original;
           if extra_metadata is not None, no extra metadata is inherited from original
        """

        if extra_metadata is None:
            extra_metadata = original.extra_metadata

        polyline = cls(original.model,
                       set_crs = original.crs_uuid,
                       set_bool = original.isclosed,
                       title = title if title else original.title,
                       originator = originator if originator else original.originator,
                       extra_metadata = extra_metadata)

        o_centre = original.balanced_centre()
        polyline.coordinates = scaling * (original.coordinates - o_centre) + o_centre
        polyline.nodepatch = (0, len(polyline.coordinates))

        return polyline

    @classmethod
    def from_trimmed_polyline(cls,
                              original,
                              start_seg,
                              end_seg,
                              start_xyz = None,
                              end_xyz = None,
                              title = None,
                              originator = None,
                              extra_metadata = None):
        """Returns a trimmed version of the original polyline.

        arguments:
           original (Polyline): the polyline from which the new polyline will be spawned
           start_seg (int): the index of the first segment in original to be kept
           end_seg (int): the index of the last segment in original to be kept
           start_xyz (triple float, optional): the new start point; if None, start of start_seg is used
           end_xyz (triple float, optional): the new end point; if None, end of end_seg is used
           title (str, optional): the citation title for the new polyline; inherited from
              original if None
           originator (str, optional): the name of the person creating the polyline; inherited
              from original if None
           extra_metadata (dict, optional): extra metadata for the new polyline; inherited from
              original if None

        returns:
           a new Polyline
        """

        assert 0 <= start_seg <= end_seg
        assert end_seg < len(original.coordinates) - 1

        if extra_metadata is None:
            extra_metadata = original.extra_metadata

        polyline = cls(original.model,
                       set_crs = original.crs_uuid,
                       set_bool = False,
                       title = title if title else original.title,
                       originator = originator if originator else original.originator,
                       extra_metadata = extra_metadata)

        polyline.coordinates = original.coordinates[start_seg:end_seg + 2].copy()
        if start_xyz is not None:
            polyline.coordinates[0, :len(start_xyz)] = start_xyz
        if end_xyz is not None:
            polyline.coordinates[-1, :len(end_xyz)] = end_xyz
        polyline.nodepatch = (0, len(polyline.coordinates))

        return polyline

    @classmethod
    def for_regular_polygon(cls, model, n, radius, centre_xyz, crs_uuid, title):
        """`Returns a closed polyline representing a regular polygon in xy plane.

        arguments:
           model (Model): the model for which the new polyline is intended
           n (int): number of sides for the regular polygon
           radius (float): distance from centre of polygon to vertices; units are crs xy unites
           centre_xyz (triple float): the centre of the polygon
           crs_uuid (UUID): the uuid of the crs for the centre point and the returned polyline
           title (str): the citation title for the new polyline

        returns:
           a new closed Polyline representing a regular polygon in the xy plane

        notee:
           z values are all set to the z value of the centre point;
           one vertex will have an x value identical to the centre and a positive y offset (due north usually);
           this method does not write to hdf5 nor create xml for the new polyline
        """

        assert n >= 3 and radius > 0.0 and len(centre_xyz) == 3 and crs_uuid is not None

        coords = np.zeros((n, 3), dtype = float)
        for i in range(n):
            theta = i * 2.0 * maths.pi / float(n)
            coords[i] = np.array((radius * maths.sin(theta), radius * maths.cos(theta), 0.0), dtype = float) +  \
                        np.array(centre_xyz, dtype = float)
        polyline = cls(model, set_bool = True, set_coord = coords, set_crs = crs_uuid, title = title)

        return polyline

    @classmethod
    def convex_hull_from_closed_polyline(cls, original, title, mode = 'crossing'):
        """Returns a new closed convex polyline being the convex hull of an original closed polyline."""

        assert original.isclosed
        assert len(original.coordinates) >= 3
        if not title:
            title = f'{original.title} convex hull'

        if len(original.coordinates) == 3 or original.is_convex():
            coords = original.coordinates
        else:
            oc = original.coordinates
            while True:
                ol = len(oc)
                assert ol >= 3
                hull_mask = np.zeros(ol, dtype = bool)
                for i in range(ol):
                    if i == 0:
                        coords = oc[1:]
                    elif i == ol - 1:
                        coords = oc[:-1]
                    else:
                        coords = np.concatenate((oc[:i], oc[i + 1:]))
                    if mode == 'crossing':
                        inside = pip.pip_cn(oc[i], coords)
                    else:
                        inside = pip.pip_wn(oc[i], coords)
                    hull_mask[i] = not inside
                if np.all(hull_mask):
                    coords = oc
                    break
                oc = oc[hull_mask]
            assert len(coords) >= 3

        polyline = cls(original.model, set_bool = True, set_coord = coords, set_crs = original.crs_uuid, title = title)
        assert polyline.is_convex(trust_metadata = False)

        return polyline

    def is_convex(self, trust_metadata = True):
        """Returns True if the polyline is closed and convex in the xy plane, otherwise False."""

        if not self.isclosed:
            return False
        if trust_metadata and self.extra_metadata is not None and 'is_convex' in self.extra_metadata.keys():
            return self.extra_metadata['is_convex'].lower() == 'true'
        nodes = len(self.coordinates)
        extras = {}
        if nodes < 3:
            result = False
        else:
            cw_bool = None
            result = True
            for node in range(nodes):
                cw = vu.clockwise(self.coordinates[node - 2], self.coordinates[node - 1], self.coordinates[node])
                if cw == 0.0:
                    continue  # striaght line section
                if cw_bool is None:
                    cw_bool = (cw > 0.0)
                elif cw_bool != (cw > 0.0):
                    result = False
                    break
            if cw_bool is None:
                result = False  # whole polyline lies in a straight line
            if result:
                extras['is_clockwise'] = str(cw_bool).lower()
        extras['is_convex'] = str(result).lower()
        self.append_extra_metadata(extras)
        return result

    def is_clockwise(self, trust_metadata = True):
        """Returns True if first non-straight triplet of nodes is clockwise in the xy plane; False if anti-clockwise.

        note:
           this method currently assumes that the xy axes are left-handed
        """

        if trust_metadata and self.extra_metadata is not None and 'is_clockwise' in self.extra_metadata.keys():
            return str(self.extra_metadata['is_clockwise']).lower() == 'true'
        result = None
        for node in range(3, len(self.coordinates)):
            cw = vu.clockwise(self.coordinates[node - 2], self.coordinates[node - 1], self.coordinates[node])
            if cw == 0.0:
                continue  # striaght line section
            result = (cw > 0.0)
            break
        if result is not None:
            self.append_extra_metadata({'is_clockwise': str(result).lower()})
        return result

    def point_is_inside_xy(self, p, mode = 'crossing'):
        """Returns True if point p is inside closed polygon, in xy plane, otherwise False.

        :meta common:
        """

        assert mode in ['crossing', 'winding'], 'unrecognised point inclusion mode: ' + str(mode)
        assert self.isclosed, 'point inclusion is not applicable to unclosed polylines'

        if mode == 'crossing':
            return pip.pip_cn(p, self.coordinates)
        return pip.pip_wn(p, self.coordinates)

    def points_are_inside_xy(self, p_array):
        """Returns bool array, True where p is inside closed polygon, in xy plane, otherwise False.

        arguments:
           p_array (numpy float array of shape (..., 2 or 3)): an array of points, each of which is
              tested for inclusion against the closed polygon

        returns:
           numpy bool array of shape p_array.shape[:-1], set True for those points which are inside
           the polygon

        :meta common:
        """

        assert self.isclosed, 'point inclusion is not applicable to unclosed polylines'
        return pip.pip_array_cn(p_array, self.coordinates)

    def segment_length(self, segment_index, in_xy = False):
        """Returns the naive length (ie.

        assuming x,y & z units are the same) of an individual segment of the polyline.
        """

        successor = self._successor(segment_index)
        d = 2 if in_xy else 3
        return vu.naive_length(self.coordinates[successor, :d] - self.coordinates[segment_index, :d])

    def segment_midpoint(self, segment_index):
        """Returns the midpoint of an individual segment of the polyline."""

        successor = self._successor(segment_index)
        return 0.5 * (self.coordinates[segment_index] + self.coordinates[successor])

    def segment_normal(self, segment_index):
        """For a closed polyline return a unit vector giving the 2D (xy) direction of an outward facing normal to a segment."""

        successor = self._successor(segment_index)
        segment_vector = self.coordinates[successor, :2] - self.coordinates[segment_index, :2]
        segment_vector = vu.unit_vector(segment_vector)
        normal_vector = np.zeros(3)
        normal_vector[0] = -segment_vector[1]
        normal_vector[1] = segment_vector[0]
        cw = self.is_clockwise()
        assert cw is not None, 'polyline is straight'
        if not cw:
            normal_vector = -normal_vector
        return normal_vector

    def full_length(self, in_xy = False):
        """Returns the naive length of the entire polyline.

        :meta common:
        """

        length = 0.0
        end_index = len(self.coordinates) - 1
        if self.isclosed:
            end_index += 1
        for i in range(end_index):
            length += self.segment_length(i, in_xy = in_xy)
        return length

    def interpolated_point(self, fraction, in_xy = False):
        """Returns x,y,z point on the polyline at fractional distance along entire polyline.

        :meta common:
        """

        assert 0.0 <= fraction <= 1.0
        target = fraction * self.full_length(in_xy = in_xy)
        seg_index = 0
        while True:
            next_seg_length = self.segment_length(seg_index, in_xy = in_xy)
            if target > next_seg_length or next_seg_length == 0.0:
                target -= next_seg_length
                seg_index += 1
                continue
            seg_fraction = target / next_seg_length
            successor = (seg_index + 1) % len(self.coordinates)
            return seg_fraction * self.coordinates[successor] + (1.0 - seg_fraction) * self.coordinates[seg_index]

    def equidistant_points(self, n, in_xy = False):
        """Returns array of shape (n, 3) being points equally distributed along entire polyline."""

        assert n > 1
        sample_points = np.empty((n, 3))
        if self.isclosed:
            step_fraction = 1.0 / float(n)
        else:
            step_fraction = 1.0 / float(n - 1)
        f = 0.0
        for step in range(n):
            sample_points[step] = self.interpolated_point(f, in_xy = in_xy)
            f += step_fraction
            if f > 1.0:
                f = 1.0  # to avoid possible assertion error due to rounding
        return sample_points

    def balanced_centre(self, mode = 'weighted', n = 20, cache = True, in_xy = False):
        """Returns a mean x,y,z based on sampling polyline at regular intervals."""

        if cache and self.centre is not None:
            return self.centre
        assert mode in ['weighted', 'sampled']
        if mode == 'sampled':  # this mode is deprecated as it simply approximates the weighted mode
            sample_points = self.equidistant_points(n, in_xy = in_xy)
            centre = np.mean(sample_points, axis = 0)
        else:  # 'weighted'
            sum = np.zeros(3)
            seg_count = len(self.coordinates) - 1
            if self.isclosed:
                seg_count += 1
            d = 2 if in_xy else 3
            p1 = np.zeros(3)
            p2 = np.zeros(3)
            for seg_index in range(seg_count):
                successor = (seg_index + 1) % len(self.coordinates)
                p1[:d], p2[:d] = self.coordinates[seg_index, :d], self.coordinates[successor, :d]
                sum += (p1 + p2) * vu.naive_length(p2 - p1)
            centre = sum / (2.0 * self.full_length(in_xy = in_xy))
        if cache:
            self.centre = centre
        return centre

    def first_line_intersection(self, x1, y1, x2, y2, half_segment = False):
        """Finds the first intersection of (half) bounded line x,y 1 to 2 with polyline.

        returns:
           segment number & x, y of first intersection of (half) bounded line x,y 1 to 2 with polyline,
           or None, None, None if no intersection found

        note:
           'first' primariliy refers to the ordering of segments in this polyline
        """

        seg_count = len(self.coordinates) - 1
        if self.isclosed:
            seg_count += 1

        for segment in range(seg_count):
            successor = (segment + 1) % len(self.coordinates)
            x, y = meet.line_line_intersect(x1,
                                            y1,
                                            x2,
                                            y2,
                                            self.coordinates[segment][0],
                                            self.coordinates[segment][1],
                                            self.coordinates[successor][0],
                                            self.coordinates[successor][1],
                                            line_segment = True,
                                            half_segment = half_segment)
            if x is not None:
                return segment, x, y
        return None, None, None

    def closest_segment_and_distance_to_point_xy(self, p):
        """Returns the index of the closest segment to a point, and its distance, in the xy plane.

        arguments:
            p (pair or triple float): the point

        returns:
            (int, float) where the int is the index of the line segment that the point is closest to;
            and the float is the distance of the point from that bounded segment, in the xy plane;
            units of measure are the crs xy units
        """

        # log.debug(f'{self.title}: closest_segment_and_distance_to_point_xy: {p}')
        # log.debug(f'{self.coordinates}')
        p = np.array(p[:2], dtype = float)
        min_distance = vu.point_distance_to_line_segment_2d(p, self.coordinates[0, :2], self.coordinates[1, :2])
        min_segment = 0
        # log.debug(f'.min_seg: {min_segment}; min_distance: {min_distance}')
        for seg in range(1, len(self.coordinates) - 1):
            distance = vu.point_distance_to_line_segment_2d(p, self.coordinates[seg, :2], self.coordinates[seg + 1, :2])
            if distance < min_distance:
                min_distance = distance
                min_segment = seg
            # log.debug(f'..min_seg: {min_segment}; min_distance: {min_distance}')
        if self.isclosed:
            distance = vu.point_distance_to_line_segment_2d(p, self.coordinates[-1, :2], self.coordinates[0, :2])
            if distance < min_distance:
                min_distance = distance
                min_segment = len(self.coordinates) - 1
            # log.debug(f'...min_seg: {min_segment}; min_distance: {min_distance}')
        # log.debug(f'min_seg: {min_segment}; min_distance: {min_distance}')
        return min_segment, min_distance

    def point_snapped_to_segment_xy(self, segment, p):
        """Returns the point on a specified segment, in xy plane, that is closest to a point.

        arguments:
            segment (int): the index of the line segment within the polyline
            p (pair or triple float): the point p (z value is ignored if present)

        returns:
            numpy float array of shape (2,) being the x, y coordinates of the snapped point
        """

        if segment == len(self.coordinates) - 1:
            segment = -1
        return meet.point_snapped_to_line_segment_2d(p, self.coordinates[segment], self.coordinates[segment + 1])

    def xy_crossings(self, other):
        """Returns list of (x, y) pairs of crossing points with other polyline, in xy plane."""

        seg_count = len(self.coordinates) - 1
        if self.isclosed:
            seg_count += 1
        other_seg_count = len(other.coordinates) - 1
        if other.isclosed:
            other_seg_count += 1

        crossings = []
        for i in range(seg_count):
            ip = (i + 1) % len(self.coordinates)
            for j in range(other_seg_count):
                jp = (j + 1) % len(other.coordinates)
                x, y = meet.line_line_intersect(self.coordinates[i, 0],
                                                self.coordinates[i, 1],
                                                self.coordinates[ip, 0],
                                                self.coordinates[ip, 1],
                                                other.coordinates[j, 0],
                                                other.coordinates[j, 1],
                                                other.coordinates[jp, 0],
                                                other.coordinates[jp, 1],
                                                line_segment = True,
                                                half_segment = False)
                if x is not None and (not crossings or
                                      not (maths.isclose(x, crossings[-1][0]) and maths.isclose(y, crossings[-1][1]))):
                    crossings.append((x, y))

        return crossings

    def normalised_xy(self, x, y, mode = 'square'):
        """Returns a normalised x',y' pair (in range 0..1) being point x,y under mapping from convex polygon.

        arguments:
            x, y (floats): location of a point inside the polyline, which must be closed and project to a
                convex polygon in the xy plane
            mode (string): which mapping algorithm to use, one of: 'square', 'circle', or 'perimeter'

        returns:
            x', y' (floats, each in range 0..1) being the normalised representation of point x,y

        notes:
            this method is the inverse of denormalised_xy(), as long as a consistent mode is selected;
            for more details of the mapping used by the 3 modes, see documentation for denormalised_xy()
        """

        assert mode in ['square', 'perimeter', 'circle']
        assert self.is_convex(), 'attempt to find normalised x,y within a polyline that is not a closed convex polygon'
        centre_xy = self.balanced_centre()[:2]
        if tuple(centre_xy) == (x, y):
            if mode == 'square':
                return 0.5, 0.5
            return 0.0, 0.0
        segment, px, py = self.first_line_intersection(centre_xy[0], centre_xy[1], x, y, half_segment = True)
        assert px is not None
        norm_x, norm_y = None, None
        if mode == 'square':  # todo: check square mode – looks wrong
            if px == centre_xy[0]:
                norm_x = 0.5
            else:
                norm_x = 0.5 + 0.5 * (x - centre_xy[0]) / abs(px - centre_xy[0])
            if py == centre_xy[1]:
                norm_y = 0.5
            else:
                norm_y = 0.5 + 0.5 * (y - centre_xy[1]) / abs(py - centre_xy[1])
        elif mode == 'circle':
            d_xy = np.array((x, y)) - centre_xy
            dp_xy = np.array((px, py)) - centre_xy
            if abs(dp_xy[0]) > abs(dp_xy[1]):
                rf = d_xy[0] / dp_xy[0]
                theta = maths.atan(d_xy[1] / d_xy[0])
                if d_xy[0] < 0.0:
                    theta += maths.pi
            else:
                rf = d_xy[1] / dp_xy[1]
                theta = maths.pi / 2.0 - maths.atan(d_xy[0] / d_xy[1])
                if d_xy[1] < 0.0:
                    theta += maths.pi
            norm_x = rf * rf
            theta /= 2.0 * maths.pi
            if theta < 0.0:
                theta += 1.0
            elif theta > 1.0:
                theta -= 1.0
            norm_y = theta
        elif mode == 'perimeter':
            d_xy = np.array((x, y)) - centre_xy
            dp_xy = np.array((px, py)) - centre_xy
            if abs(dp_xy[0]) > abs(dp_xy[1]):
                rf = d_xy[0] / dp_xy[0]
            else:
                rf = d_xy[1] / dp_xy[1]
            norm_x = rf * rf
            seg_xy = np.array((px, py)) - self.coordinates[segment, :2]
            length = 0.0
            for seg in range(segment):
                length += self.segment_length(seg, in_xy = True)
            length += vu.naive_length(seg_xy)
            norm_y = length / self.full_length(in_xy = True)
        return norm_x, norm_y

    def denormalised_xy(self, norm_x, norm_y, mode = 'perimeter'):
        """Returns a denormalised x,y pair being point norm_x,norm_y (in range 0..1) under mapping onto convex polygon.

        arguments:
            norm_x, norm_y (floats): normalised values, each in range 0..1, identifying a location in a unit shape
            mode (string): 'square', 'circle', or 'perimeter'; if square, norm_x and norm_y are coordinates within a
                unit square; if circle or perimeter, norm_x is the square of a radial fraction between the centre
                of the polygon and the perimeter; if mode is circle, norm_y is simply the polar coordinate bearing
                of the point relative to the polygon centre, with values 0..1 corresponding to 0..2pi radians;
                if mode is perimeter, norm_y is a fractional distance along the length of the closed polyline
                projected onto the xy plane

        returns:
            x, y (floats): the location of a point within the area outlined by the polyline, which must be closed
                and project to a convex polygon in the xy plane

        notes:
            this method is the inverse of normalised_xy(), as long as a consistent mode is selected;
            each mode gives a reversible mapping but there are some variations in density which could introduce
            slightly different biases when working with stochastic locations; density distortion with square and
            circle modes tends to increase as the polygon shape becomes less square (and aligned) or circular
            respectively; perimeter mode is believed to yield least density distortion generally;
            for circle and perimeter modes, the normalised x value is treated as the square of the fractional
            distance to the boundary in order for a rectangular distribution of normalised values to map to
            an even density of denormalised locations
        """

        assert mode in ['square', 'perimeter', 'circle']
        assert 0.0 <= norm_x <= 1.0 and 0.0 <= norm_y <= 1.0
        assert self.is_convex(
        ), 'attempt to find denormalised x,y within a polyline that is not a closed convex polygon'
        centre_xy = self.balanced_centre()[:2]
        x, y = None, None
        if mode == 'square':
            if (norm_x, norm_y) == (0.5, 0.5):
                return tuple(centre_xy)
            n_x = (norm_x - 0.5) + centre_xy[0]
            n_y = (norm_y - 0.5) + centre_xy[1]
            _, px, py = self.first_line_intersection(centre_xy[0], centre_xy[1], n_x, n_y, half_segment = True)
            assert px is not None
            fx = abs(norm_x - 0.5)
            fy = abs(norm_y - 0.5)
            f = 2.0 * max(fx, fy)
            # log.debug(f'intersect x,y: {px}, {py}')
            if px == centre_xy[0]:
                x = centre_xy[0]
            else:
                x = centre_xy[0] + f * (px - centre_xy[0])
            if py == centre_xy[1]:
                y = centre_xy[1]
            else:
                y = centre_xy[1] + f * (py - centre_xy[1])
            # log.debug(f'denormal x,y: {x}, {y}')
        elif mode == 'perimeter':
            if norm_y == 1.0:
                norm_y = 0.0
            p_xy = np.empty(2)
            p_xy[:] = self.interpolated_point(norm_y)[:2]
            f = maths.sqrt(norm_x)
            x, y = centre_xy + f * (p_xy - centre_xy)
        elif mode == 'circle':
            theta = 2 * maths.pi * norm_y
            n_x = centre_xy[0] + maths.cos(theta)
            n_y = centre_xy[1] + maths.sin(theta)
            p_xy = np.empty(2)
            p_xy[:] = self.first_line_intersection(centre_xy[0], centre_xy[1], n_x, n_y, half_segment = True)[1:]
            f = maths.sqrt(norm_x)
            x, y = centre_xy + f * (p_xy - centre_xy)
        return x, y

    def tangent_vectors(self):
        """Returns a numpy array of unit length tangent vectors, one for each coordinate in the line."""

        return rql_c.tangents(self.coordinates)

    def splined(self,
                tangent_weight = 'square',
                min_subdivisions = 1,
                max_segment_length = None,
                max_degrees_per_knot = 5.0,
                title = None,
                rep_int_root = None):
        """Retrurns a new Polyline being a cubic spline of this polyline.

        :meta common:
        """

        spline_coords = rql_c.spline(self.coordinates,
                                     tangent_weight = tangent_weight,
                                     min_subdivisions = min_subdivisions,
                                     max_segment_length = max_segment_length,
                                     max_degrees_per_knot = max_degrees_per_knot,
                                     closed = self.isclosed)

        if not title:
            title = self.title
        if rep_int_root is None:
            rep_int_root = self.rep_int_root  # todo: check whether it is legal to have 2 representations for 1 interpretation

        return Polyline(self.model,
                        set_bool = self.isclosed,
                        set_coord = spline_coords,
                        set_crs = self.crs_uuid,
                        title = title,
                        rep_int_root = rep_int_root)

    def area(self):
        """Returns the area in the xy plane of a closed convex polygon."""
        assert self.isclosed
        if self.is_convex():
            centre = np.mean(self.coordinates, axis = 0)
            a = 0.0
            for node in range(len(self.coordinates)):
                a += vu.area_of_triangle(centre, self.coordinates[node - 1], self.coordinates[node])
        else:  # use a regular 2D sampling of points to determine approx area
            xy_box = np.zeros((2, 2), dtype = float)
            xy_box[0] = np.min(self.coordinates[:, :2], axis = 0)
            xy_box[1] = np.max(self.coordinates[:, :2], axis = 0)
            assert np.all(xy_box[1] > xy_box[0])
            d_xy = xy_box[1] - xy_box[0]
            d = d_xy[0] + d_xy[1]
            f_xy = d_xy / d
            n_xy = np.ceil(100.0 * f_xy).astype(int)
            x = np.linspace(xy_box[0, 0], xy_box[1, 0], num = n_xy[0], dtype = float)
            y = np.linspace(xy_box[0, 1], xy_box[1, 1], num = n_xy[1], dtype = float)
            xy = np.stack(np.meshgrid(x, y), axis = -1).reshape((-1, 2))
            inside = pip.pip_array_cn(xy, self.coordinates)
            nf_xy = n_xy.astype(float)
            d_xy *= (nf_xy + 1.0) / nf_xy
            a = d_xy[0] * d_xy[1] * float(np.count_nonzero(inside)) / float(inside.size)
        return a

    def create_xml(self,
                   ext_uuid = None,
                   add_as_part = True,
                   add_relationships = True,
                   title = None,
                   originator = None):
        """Create xml from polyline.

        args:
            ext_uuid: the uuid of the hdf5 external part

        :meta common:
        """

        if ext_uuid is None:
            ext_uuid = self.model.h5_uuid()

        if title is not None:
            self.title = title
        if self.title is None:
            self.title = 'polyline'

        polyline = super().create_xml(add_as_part = False, originator = originator)

        if self.rep_int_root is not None:
            rep_int = self.rep_int_root
            if "FaultInterpretation" in str(rqet.content_type(rep_int)):
                content_type = 'obj_FaultInterpretation'
            else:
                content_type = 'obj_HorizonInterpretation'
            self.model.create_ref_node('RepresentedInterpretation',
                                       rqet.citation_title_for_node(rep_int),
                                       rep_int.attrib['uuid'],
                                       content_type = content_type,
                                       root = polyline)

        isclosed = rqet.SubElement(polyline, ns['resqml2'] + 'IsClosed')
        isclosed.set(ns['xsi'] + 'type', ns['xsd'] + 'boolean')
        isclosed.text = str(self.isclosed).lower()

        nodepatch = rqet.SubElement(polyline, ns['resqml2'] + 'NodePatch')
        nodepatch.set(ns['xsi'] + 'type', ns['resqml2'] + 'NodePatch')
        nodepatch.text = '\n'

        patchindex = rqet.SubElement(nodepatch, ns['resqml2'] + 'PatchIndex')
        patchindex.set(ns['xsi'] + 'type', ns['xsd'] + 'nonNegativeInteger')
        patchindex.text = str(self.nodepatch[0])

        count = rqet.SubElement(nodepatch, ns['resqml2'] + 'Count')
        count.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
        count.text = str(self.nodepatch[1])

        geom = rqet.SubElement(nodepatch, ns['resqml2'] + 'Geometry')
        geom.set(ns['xsi'] + 'type', ns['resqml2'] + 'PointGeometry')
        geom.text = '\n'

        self.model.create_crs_reference(crs_uuid = self.crs_uuid, root = geom)

        points = rqet.SubElement(geom, ns['resqml2'] + 'Points')
        points.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dHdf5Array')
        points.text = '\n'

        coords = rqet.SubElement(points, ns['resqml2'] + 'Coordinates')
        coords.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
        coords.text = rqet.null_xml_text

        self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'points_patch0', root = coords)

        if add_as_part:
            self.model.add_part('obj_PolylineRepresentation', self.uuid, polyline)
            if add_relationships:
                crs_root = self.model.root_for_uuid(self.crs_uuid)
                self.model.create_reciprocal_relationship(polyline, 'destinationObject', crs_root, 'sourceObject')
                if self.rep_int_root is not None:  # Optional
                    self.model.create_reciprocal_relationship(polyline, 'destinationObject', self.rep_int_root,
                                                              'sourceObject')

                ext_part = rqet.part_name_for_object('obj_EpcExternalPartReference', ext_uuid, prefixed = False)
                ext_node = self.model.root_for_part(ext_part)
                self.model.create_reciprocal_relationship(polyline, 'mlToExternalPartProxy', ext_node,
                                                          'externalPartProxyToMl')

        return polyline

    def write_hdf5(self, file_name = None, mode = 'a'):
        """Create or append the coordinates hdf5 array to hdf5 file.

        :meta common:
        """

        if self.uuid is None:
            self.uuid = bu.new_uuid()
        h5_reg = rwh5.H5Register(self.model)
        h5_reg.register_dataset(self.uuid, 'points_patch0', self.coordinates)
        h5_reg.write(file_name, mode = mode)

    def _successor(self, segment_index):
        """Returns segment_index + 1; or zero if segment_index is the last segment in a closed polyline."""

        successor = segment_index + 1
        if self.isclosed:
            assert 0 <= segment_index < len(self.coordinates)
            if segment_index == len(self.coordinates) - 1:
                successor = 0
        else:
            assert 0 <= segment_index < len(self.coordinates) - 1
        return successor



"""TimeSeries class handling normal (non-geological) time series."""

import logging

log = logging.getLogger(__name__)

import datetime as dt
import warnings

import resqpy.time_series
import resqpy.time_series._any_time_series as ats
import resqpy.time_series._time_duration as td


class TimeSeries(ats.AnyTimeSeries):
    """Class for RESQML Time Series without year offsets.

    notes:
       use this class for time series on a human timeframe; use the resqpy GeologicTimeSeries class
       instead if the time series is on a geological timeframe
    """

    def __init__(self,
                 parent_model,
                 uuid = None,
                 first_timestamp = None,
                 daily = None,
                 monthly = None,
                 quarterly = None,
                 yearly = None,
                 title = None,
                 originator = None,
                 extra_metadata = None):
        """Create a TimeSeries object, either from a time series node in parent model, or from given data.

        arguments:
           parent_model (model.Model): the resqpy model to which the time series will belong
           uuid (uuid.UUID, optional): the uuid of a TimeSeries object to be loaded from xml
           first_time_stamp (str, optional): the first timestamp (in RESQML format) if not loading from xml;
              this and the remaining arguments are ignored if loading from xml
           daily (non-negative int, optional): the number of one day interval timesteps to start the series
           monthly (non-negative int, optional): the number of 30 day interval timesteps to follow the daily
              timesteps
           quarterly (non-negative int, optional): the number of 90 day interval timesteps to follow the
              monthly timesteps
           yearly (non-negative int, optional): the number of 365 day interval timesteps to follow the
              quarterly timesteps
           title (str, optional): the citation title to use for a new time series;
              ignored if uuid is not None
           originator (str, optional): the name of the person creating the time series, defaults to login id;
              ignored if uuid is not None
           extra_metadata (dict, optional): string key, value pairs to add as extra metadata for the time series;
              ignored if uuid is not None

        returns:
           newly instantiated TimeSeries object

        note:
           a new bespoke time series can be populated by passing the first timestamp here and using the
           add_timestamp() and/or extend_by...() methods

        :meta common:
        """

        self.timeframe = 'human'
        self.timestamps = []  # ordered list of timestamp strings in resqml/iso format
        if first_timestamp is not None:
            self.timestamps.append(first_timestamp)  # todo: check format of first_timestamp
            if daily is not None:
                for _ in range(daily):
                    self.extend_by_days(1)
            if monthly is not None:
                for _ in range(monthly):
                    self.extend_by_days(30)
            if quarterly is not None:
                for _ in range(quarterly):
                    self.extend_by_days(90)  # could use 91
            if yearly is not None:
                for _ in range(yearly):
                    self.extend_by_days(365)  # could use 360
        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)
        if self.extra_metadata is not None and self.extra_metadata.get('timeframe') == 'geologic':
            raise ValueError('attempt to instantiate a human timeframe time series for a geologic time series')

    def is_equivalent(self, other_ts, tol_seconds = 1):
        """Returns True if the this timestep series is essentially identical to the other; otherwise False."""

        super_equivalence = super().is_equivalent(other_ts)
        if super_equivalence is not None:
            return super_equivalence
        tolerance = td.TimeDuration(seconds = tol_seconds)
        for t_index in range(self.number_of_timestamps()):
            diff = td.TimeDuration(earlier_timestamp = self.timestamps[t_index],
                                   later_timestamp = other_ts.timestamps[t_index])
            if abs(diff.duration) > tolerance.duration:
                return False
        return True

    def index_for_timestamp_not_later_than(self, timestamp):
        """Returns the index of the latest timestamp that is not later than the specified date.

        :meta common:
        """
        index = len(self.timestamps) - 1
        while (index >= 0) and (self.timestamps[index] > timestamp):
            index -= 1
        if index < 0:
            return None
        return index

    def index_for_timestamp_not_earlier_than(self, timestamp):
        """Returns the index of the earliest timestamp that is not earlier than the specified date.

        :meta common:
        """
        index = 0
        while (index < len(self.timestamps)) and (self.timestamps[index] < timestamp):
            index += 1
        if index >= len(self.timestamps):
            return None
        return index

    def index_for_timestamp_closest_to(self, timestamp):
        """Returns the index of the timestamp that is closest to the specified date.

        :meta common:
        """
        if not self.timestamps:
            return None
        before = self.index_for_timestamp_not_later_than(timestamp)
        if not before:
            return 0
        if before == len(self.timestamps) - 1 or self.timestamps[before] == timestamp:
            return before
        after = before + 1
        early_delta = td.TimeDuration(earlier_timestamp = self.timestamps[before], later_timestamp = timestamp)
        later_delta = td.TimeDuration(earlier_timestamp = timestamp, later_timestamp = self.timestamps[after])
        return before if early_delta.duration <= later_delta.duration else after

    def duration_between_timestamps(self, earlier_index, later_index):
        """Returns the duration between a pair of timestamps.

        :meta common:
        """
        if earlier_index < 0 or later_index >= len(self.timestamps) or later_index < earlier_index:
            return None
        return td.TimeDuration(earlier_timestamp = self.timestamps[earlier_index],
                               later_timestamp = self.timestamps[later_index])

    def days_between_timestamps(self, earlier_index, later_index):
        """Returns the number of whole days between a pair of timestamps, as an integer."""
        delta = self.duration_between_timestamps(earlier_index, later_index)
        if delta is None:
            return None
        return delta.duration.days

    def duration_since_start(self, index):
        """Returns the duration between the start of the time series and the indexed timestamp.

        :meta common:
        """
        if index < 0 or index >= len(self.timestamps):
            return None
        return self.duration_between_timestamps(0, index)

    def days_since_start(self, index):
        """Returns the number of days between the start of the time series and the indexed timestamp."""
        return self.duration_since_start(index).duration.days

    def step_duration(self, index):
        """Returns the duration of the time step between the indexed timestamp and preceding one.

        :meta common:
        """
        if index < 1 or index >= len(self.timestamps):
            return None
        return self.duration_between_timestamps(index - 1, index)

    def step_days(self, index):
        """Returns the number of days between the indexed timestamp and preceding one."""
        delta = self.step_duration(index)
        if delta is None:
            return None
        return delta.duration.days

    # NB: Following functions modify the time series, which is dangerous if the series is in use by a model
    # Could check for relationships involving the time series and disallow changes if any found?
    def add_timestamp(self, new_timestamp, allow_insertion = False):
        """Inserts a new timestamp into the time series."""
        # todo: check that new_timestamp is in valid format (iso format + 'Z')
        if allow_insertion:
            # NB: This can insert a timestamp anywhere in the series, will invalidate indices, possibly corrupting model
            index = self.index_for_timestamp_not_later_than(new_timestamp)
            if index is None:
                index = 0
            else:
                index += 1
            self.timestamps.insert(index, new_timestamp)
        else:
            last = self.last_timestamp()
            if last is not None:
                assert (new_timestamp > self.last_timestamp())
            self.timestamps.append(new_timestamp)

    def extend_by_duration(self, duration):
        """Adds a timestamp to the end of the series, at duration beyond the last timestamp."""
        assert (duration.duration.days >= 0)  # duration may not be negative
        assert (len(self.timestamps) > 0)  # there must be something to extend from
        self.timestamps.append(duration.timestamp_after_duration(self.last_timestamp()))

    def extend_by_days(self, days):
        """Adds a timestamp to the end of the series, at a duration of days beyond the last timestamp."""
        duration = td.TimeDuration(days = days)
        self.extend_by_duration(duration)

    def datetimes(self):
        """Returns the timestamps as a list of python-datetime objects."""
        return [dt.datetime.fromisoformat(t.rstrip('Z')) for t in self.timestamps]



"""Submodule containing the face related grid functions."""

import logging

log = logging.getLogger(__name__)
import numpy as np
import pandas as pd

import resqpy.fault as rqf
import resqpy.olio.vector_utilities as vec


def is_split_column_face(grid, j0, i0, axis, polarity):
    """Returns True if the I or J column face is split; False otherwise."""

    if not grid.has_split_coordinate_lines:
        return False
    assert axis in (1, 2)
    if axis == 1:  # J
        ip = i0
        if polarity:
            if j0 == grid.nj - 1:
                return False
            jp = j0 + 1
        else:
            if j0 == 0:
                return False
            jp = j0 - 1
    else:  # I
        jp = j0
        if polarity:
            if i0 == grid.ni - 1:
                return False
            ip = i0 + 1
        else:
            if i0 == 0:
                return False
            ip = i0 - 1
    cpm = grid.create_column_pillar_mapping()
    if axis == 1:
        return ((cpm[j0, i0, polarity, 0] != cpm[jp, ip, 1 - polarity, 0]) or
                (cpm[j0, i0, polarity, 1] != cpm[jp, ip, 1 - polarity, 1]))
    else:
        return ((cpm[j0, i0, 0, polarity] != cpm[jp, ip, 0, 1 - polarity]) or
                (cpm[j0, i0, 1, polarity] != cpm[jp, ip, 1, 1 - polarity]))


def split_column_faces(grid):
    """Returns a pair of numpy boolean arrays indicating which internal column faces (column edges) are split."""

    if not grid.has_split_coordinate_lines:
        return None, None
    if (hasattr(grid, 'array_j_column_face_split') and grid.array_j_column_face_split is not None and
            hasattr(grid, 'array_i_column_face_split') and grid.array_i_column_face_split is not None):
        return grid.array_j_column_face_split, grid.array_i_column_face_split
    if grid.nj == 1:
        grid.array_j_column_face_split = None
    else:
        grid.array_j_column_face_split = np.zeros((grid.nj - 1, grid.ni),
                                                  dtype = bool)  # NB. internal faces only, index for +ve face
    if grid.ni == 1:
        grid.array_i_column_face_split = None
    else:
        grid.array_i_column_face_split = np.zeros((grid.nj, grid.ni - 1),
                                                  dtype = bool)  # NB. internal faces only, index for +ve face
    grid.create_column_pillar_mapping()
    for spi in grid.split_pillar_indices_cached:
        j_p, i_p = divmod(spi, grid.ni + 1)
        if j_p > 0 and j_p < grid.nj:
            if i_p > 0 and grid.is_split_column_face(j_p, i_p - 1, 1, 0):
                grid.array_j_column_face_split[j_p - 1, i_p - 1] = True
            if i_p < grid.ni - 1 and grid.is_split_column_face(j_p, i_p, 1, 0):
                grid.array_j_column_face_split[j_p - 1, i_p] = True
        if i_p > 0 and i_p < grid.ni:
            if j_p > 0 and grid.is_split_column_face(j_p - 1, i_p, 2, 0):
                grid.array_i_column_face_split[j_p - 1, i_p - 1] = True
            if j_p < grid.nj - 1 and grid.is_split_column_face(j_p, i_p, 2, 0):
                grid.array_i_column_face_split[j_p, i_p - 1] = True
    return grid.array_j_column_face_split, grid.array_i_column_face_split


def clear_face_sets(grid):
    """Discard face sets."""
    # following maps face_set_id to (j faces, i faces, 'K') of array of [j, i] kelp indices for that face_set_id
    # or equivalent for axes 'J' or 'I'
    grid.face_set_dict = {}
    grid.face_set_gcs_list = []


def set_face_set_gcs_list_from_dict(grid, face_set_dict = None, create_organizing_objects_where_needed = False):
    """Creates a grid connection set for each feature in the face set dictionary, based on kelp list pairs."""

    if face_set_dict is None:
        face_set_dict = grid.face_set_dict
    grid.face_set_gcs_list = []
    for feature, kelp_values in face_set_dict.items():
        gcs = rqf.GridConnectionSet(grid.model, grid = grid)
        if len(kelp_values) == 2:
            kelp_j, kelp_i = kelp_values
            axis = 'K'
        elif len(kelp_values) == 3:
            kelp_j, kelp_i, axis = kelp_values
        else:
            raise ValueError('grid face set dictionary item messed up')
        log.debug(f'creating gcs for: {feature} {axis}')
        gcs.set_pairs_from_kelp(kelp_j, kelp_i, feature, create_organizing_objects_where_needed, axis = axis)
        grid.face_set_gcs_list.append(gcs)


# TODO: make separate curtain and K-face versions of following function
def make_face_set_from_dataframe(grid, df):
    """Creates a curtain face set for each named fault in dataframe.

    note:
       this method is deprecated, or due for overhaul to make compatible with resqml_fault module and the
       GridConnectionSet class
    """

    # df columns: name, i1, i2, j1, j2, k1, k2, face
    grid.clear_face_sets()
    names = pd.unique(df.name)
    count = 0
    box_kji0 = np.zeros((2, 3), dtype = int)
    k_warning_given = False
    for fs_name in names:
        i_kelp_list = []
        j_kelp_list = []
        # ignore k faces for now
        fs_ds = df[df.name == fs_name]
        for row in range(len(fs_ds)):
            face = fs_ds.iloc[row]['face']
            fl = face[0].upper()
            if fl in 'IJK':
                axis = 'KJI'.index(fl)
            elif fl in 'XYZ':
                axis = 'ZYX'.index(fl)
            else:
                raise ValueError('fault data face not recognized: ' + face)
            if axis == 0:
                continue  # ignore k faces for now
            box_kji0[0, 0] = fs_ds.iloc[row]['k1'] - 1  # k1
            box_kji0[1, 0] = fs_ds.iloc[row]['k2'] - 1  # k2
            box_kji0[0, 1] = fs_ds.iloc[row]['j1'] - 1  # j1
            box_kji0[1, 1] = fs_ds.iloc[row]['j2'] - 1  # j2
            box_kji0[0, 2] = fs_ds.iloc[row]['i1'] - 1  # i1
            box_kji0[1, 2] = fs_ds.iloc[row]['i2'] - 1  # i2
            box_kji0[1, 0] = min(box_kji0[1, 0], grid.extent_kji[0] - 1)
            if not k_warning_given and (box_kji0[0, 0] != 0 or box_kji0[1, 0] != grid.extent_kji[0] - 1):
                log.warning(
                    'one or more entries in face set dataframe does not cover entire layer range: extended to all layers'
                )
                k_warning_given = True
            if len(face) > 1 and face[1] == '-':  # treat negative faces as positive faces of neighbouring cell
                box_kji0[0, axis] = max(box_kji0[0, axis] - 1, 0)
                box_kji0[1, axis] -= 1
            else:
                box_kji0[1, axis] = min(box_kji0[1, axis], grid.extent_kji[axis] - 2)
            if box_kji0[1, axis] < box_kji0[0, axis]:
                continue  # faces are all on edge of grid
            # for now ignore layer range and create curtain of j and i kelp
            for j in range(box_kji0[0, 1], box_kji0[1, 1] + 1):
                for i in range(box_kji0[0, 2], box_kji0[1, 2] + 1):
                    if axis == 1:
                        __add_to_kelp_list(grid.extent_kji, j_kelp_list, True, (j, i))
                    elif axis == 2:
                        __add_to_kelp_list(grid.extent_kji, i_kelp_list, False, (j, i))
        grid.face_set_dict[fs_name] = (j_kelp_list, i_kelp_list, 'K')
        count += 1
    log.info(str(count) + ' face sets extracted from dataframe')


def make_face_sets_from_pillar_lists(grid,
                                     pillar_list_list,
                                     face_set_id,
                                     axis = 'K',
                                     ref_slice0 = 0,
                                     plus_face = False,
                                     projection = 'xy'):
    """Creates a curtain face set for each pillar (or rod) list.

    returns:
       (face_set_dict, full_pillar_list_dict)

    note:
       'xz' and 'yz' projections currently only supported for unsplit grids
    """

    # NB. this code was originally written for axis K and projection xy, working with horizon points
    # it has since been reworked for the cross sectional cases, so variables named 'pillar...' may refer to rods
    # and i_... and j_... may actually represent i,j or i,k or j,k

    assert axis in ['K', 'J', 'I']
    assert projection in ['xy', 'xz', 'yz']

    local_face_set_dict = {}
    full_pillar_list_dict = {}

    if not hasattr(grid, 'face_set_dict'):
        grid.clear_face_sets()

    if axis.upper() == 'K':
        assert projection == 'xy'
        pillar_xy = grid.horizon_points(ref_k0 = ref_slice0, kp = 1 if plus_face else 0)[:, :, 0:2]
        kelp_axes = 'JI'
    else:
        if projection == 'xz':
            pillar_xy = grid.unsplit_x_section_points(axis, ref_slice0 = ref_slice0,
                                                      plus_face = plus_face)[:, :, 0:3:2]  # x,z
        else:  # projection == 'yz'
            pillar_xy = grid.unsplit_x_section_points(axis, ref_slice0 = ref_slice0, plus_face = plus_face)[:, :,
                                                                                                            1:]  # y,z
        if axis.upper() == 'J':
            kelp_axes = 'KI'
        else:
            kelp_axes = 'KJ'
    kelp_axes_int = np.empty((2,), dtype = int)
    for a in range(2):
        kelp_axes_int[a] = 'KJI'.index(kelp_axes[a])
    #     grid.clear_face_sets()  # now accumulating sets
    face_set_count = 0
    here = np.zeros(2, dtype = int)
    side_step = np.zeros(2, dtype = int)
    for pillar_list in pillar_list_list:
        if len(pillar_list) < 2:
            continue
        full_pillar_list = [pillar_list[0]]
        face_set_count += 1
        if len(pillar_list_list) > 1:
            id_suffix = '_line_' + str(face_set_count)
        else:
            id_suffix = ''
        # i,j are as stated if axis is 'K'; for axis 'J', i,j are actually i,k; for axis 'I', i,j are actually j,k
        i_kelp_list = []
        j_kelp_list = []
        for p in range(len(pillar_list) - 1):
            ji_0 = pillar_list[p]
            ji_1 = pillar_list[p + 1]
            if np.all(ji_0 == ji_1):
                continue
            # xy might actually be xy, xz or yz depending on projection
            xy_0 = pillar_xy[tuple(ji_0)]
            xy_1 = pillar_xy[tuple(ji_1)]
            if vec.isclose(xy_0, xy_1):
                continue
            dj = ji_1[0] - ji_0[0]
            di = ji_1[1] - ji_0[1]
            abs_dj = abs(dj)
            abs_di = abs(di)
            if dj < 0:
                j_sign = -1
            else:
                j_sign = 1
            if di < 0:
                i_sign = -1
            else:
                i_sign = 1
            here[:] = ji_0
            while np.any(here != ji_1):
                previous = here.copy()  # debug
                if abs_dj >= abs_di:
                    __j_greater_than_i(di, full_pillar_list, grid, here, i_kelp_list, i_sign, j_kelp_list, j_sign, ji_1,
                                       kelp_axes, kelp_axes_int, pillar_xy, side_step, xy_0, xy_1)
                else:
                    __i_greater_than_j(dj, full_pillar_list, grid, here, i_kelp_list, i_sign, j_kelp_list, j_sign, ji_1,
                                       kelp_axes, kelp_axes_int, pillar_xy, side_step, xy_0, xy_1)
                assert np.any(here != previous), 'failed to move'
        grid.face_set_dict[face_set_id + id_suffix] = (j_kelp_list, i_kelp_list, axis)
        local_face_set_dict[face_set_id + id_suffix] = (j_kelp_list, i_kelp_list, axis)
        full_pillar_list_dict[face_set_id + id_suffix] = full_pillar_list.copy()

    return local_face_set_dict, full_pillar_list_dict


def __i_greater_than_j(dj, full_pillar_list, grid, here, i_kelp_list, i_sign, j_kelp_list, j_sign, ji_1, kelp_axes,
                       kelp_axes_int, pillar_xy, side_step, xy_0, xy_1):
    i = here[1]
    if i != ji_1[1]:
        ip = i + i_sign
        __add_to_kelp_list(grid.extent_kji, j_kelp_list, kelp_axes[0], (here[0] - 1, min(i, ip)))
        here[1] = ip
        full_pillar_list.append(tuple(here))
    if dj != 0:
        divergence = vec.point_distance_to_line_2d(pillar_xy[tuple(here)], xy_0, xy_1)
        side_step[:] = here
        side_step[0] += j_sign
        if side_step[0] >= 0 and side_step[0] <= grid.extent_kji[kelp_axes_int[0]]:
            stepped_divergence = vec.point_distance_to_line_2d(pillar_xy[tuple(side_step)], xy_0, xy_1)
            if stepped_divergence < divergence:
                here[:] = side_step
                __add_to_kelp_list(grid.extent_kji, i_kelp_list, kelp_axes[1],
                                   (min(here[0], here[0] - j_sign), here[1] - 1))
                full_pillar_list.append(tuple(here))


def __j_greater_than_i(di, full_pillar_list, grid, here, i_kelp_list, i_sign, j_kelp_list, j_sign, ji_1, kelp_axes,
                       kelp_axes_int, pillar_xy, side_step, xy_0, xy_1):
    j = here[0]
    if j != ji_1[0]:
        jp = j + j_sign
        __add_to_kelp_list(grid.extent_kji, i_kelp_list, kelp_axes[1], (min(j, jp), here[1] - 1))
        here[0] = jp
        full_pillar_list.append(tuple(here))
    if di != 0:
        divergence = vec.point_distance_to_line_2d(pillar_xy[tuple(here)], xy_0, xy_1)
        side_step[:] = here
        side_step[1] += i_sign
        if side_step[1] >= 0 and side_step[1] <= grid.extent_kji[kelp_axes_int[1]]:
            stepped_divergence = vec.point_distance_to_line_2d(pillar_xy[tuple(side_step)], xy_0, xy_1)
            if stepped_divergence < divergence:
                here[:] = side_step
                __add_to_kelp_list(grid.extent_kji, j_kelp_list, kelp_axes[0],
                                   (here[0] - 1, min(here[1], here[1] - i_sign)))
                full_pillar_list.append(tuple(here))


def face_centre(grid,
                cell_kji0,
                axis,
                zero_or_one,
                points_root = None,
                cache_resqml_array = True,
                cache_cp_array = False):
    """Returns xyz location of the centre point of a face of the cell (or all cells)."""

    if axis not in [0, 1, 2]:
        raise ValueError('Axis provided must be either 0, 1, or 2.')

    # todo: optionally compute for all cells and cache
    cp = grid.corner_points(cell_kji0,
                            points_root = points_root,
                            cache_resqml_array = cache_resqml_array,
                            cache_cp_array = cache_cp_array)
    if cell_kji0 is None:
        if axis == 0:
            return 0.25 * np.sum(cp[:, :, :, zero_or_one, :, :], axis = (3, 4))
        elif axis == 1:
            return 0.25 * np.sum(cp[:, :, :, :, zero_or_one, :], axis = (3, 4))
        else:
            return 0.25 * np.sum(cp[:, :, :, :, :, zero_or_one], axis = (3, 4))
    else:
        if axis == 0:
            return 0.25 * np.sum(cp[zero_or_one, :, :], axis = (0, 1))
        elif axis == 1:
            return 0.25 * np.sum(cp[:, zero_or_one, :], axis = (0, 1))
        else:
            return 0.25 * np.sum(cp[:, :, zero_or_one], axis = (0, 1))


def face_centres_kji_01(grid, cell_kji0, points_root = None, cache_resqml_array = True, cache_cp_array = False):
    """Returns an array of shape (3, 2, 3) being (axis, 0 or 1, xyz) of face centre points for cell."""

    assert cell_kji0 is not None
    result = np.zeros((3, 2, 3))
    for axis in range(3):
        for zero_or_one in range(2):
            result[axis, zero_or_one] = grid.face_centre(cell_kji0,
                                                         axis,
                                                         zero_or_one,
                                                         points_root = points_root,
                                                         cache_resqml_array = cache_resqml_array,
                                                         cache_cp_array = cache_cp_array)
    return result


def __add_to_kelp_list(extent_kji, kelp_list, face_axis, ji):
    if isinstance(face_axis, bool):
        face_axis = 'J' if face_axis else 'I'
    # ignore external faces
    if face_axis == 'J':
        if ji[0] < 0 or ji[0] >= extent_kji[1] - 1:
            return
    elif face_axis == 'I':
        if ji[1] < 0 or ji[1] >= extent_kji[2] - 1:
            return
    else:  # ji is actually kj or ki
        assert face_axis == 'K'
        if ji[0] < 0 or ji[0] >= extent_kji[0] - 1:
            return
    pair = ji
    if pair in kelp_list:
        return  # avoid duplication
    kelp_list.append(pair)



"""Common functions and valid xml constant values for stratigraphy related RESQML classes."""

# note: two compositions have a spurious trailing space in the RESQML xsd; resqpy hides this from calling code
valid_compositions = [
    'intrusive clay ', 'intrusive clay', 'organic', 'intrusive mud ', 'intrusive mud', 'evaporite salt',
    'evaporite non salt', 'sedimentary siliclastic', 'carbonate', 'magmatic intrusive granitoid',
    'magmatic intrusive pyroclastic', 'magmatic extrusive lava flow', 'other chemichal rock', 'sedimentary turbidite'
]

valid_implacements = ['autochtonous', 'allochtonous']

valid_domains = ('depth', 'time', 'mixed')

valid_deposition_modes = [
    'proportional between top and bottom', 'parallel to bottom', 'parallel to top', 'parallel to another boundary'
]

valid_ordering_criteria = ['age', 'apparent depth', 'measured depth']  # stratigraphic column must be ordered by age

valid_contact_relationships = [
    'frontier feature to frontier feature', 'genetic boundary to frontier feature',
    'genetic boundary to genetic boundary', 'genetic boundary to tectonic boundary',
    'stratigraphic unit to frontier feature', 'stratigraphic unit to stratigraphic unit',
    'tectonic boundary to frontier feature', 'tectonic boundary to genetic boundary',
    'tectonic boundary to tectonic boundary'
]

valid_contact_verbs = ['splits', 'interrupts', 'contains', 'erodes', 'stops at', 'crosses', 'includes']

valid_contact_sides = ['footwall', 'hanging wall', 'north', 'south', 'east', 'west', 'younger', 'older', 'both']

valid_contact_modes = ['baselap', 'erosion', 'extended', 'proportional']


def _index_attr(obj):
    """Returns the index attribute of any object – typically used as a sort key function."""
    return obj.index



"""model.py: Main resqml interface module handling epc packing & unpacking and xml structures."""

import logging

log = logging.getLogger(__name__)

import h5py
import os
import warnings
from typing import Iterable, Optional, Union

import resqpy.model._catalogue as m_c
import resqpy.model._forestry as m_f
import resqpy.model._grids as m_g
import resqpy.model._hdf5 as m_h
import resqpy.model._xml as m_x
import resqpy.olio.xml_et as rqet
from resqpy import __version__

log.debug('resqpy Model class version: ' + __version__ + '; xml citation format: ' + m_x.citation_format)


class Model():
    """Class for RESQML (v2) based models.

    Examples:
          To open an existing dataset::

             Model(epc_file = 'filename.epc')

          To create a new, empty model ready to populate::

             Model(epc_file = 'new_file.epc', new_epc = True, create_basics = True, create_hdf5_ext = True)

          Alternatively, use the module level convenience function::

             new_model(epc_file = 'new_file.epc')

          To copy an existing dataset then open the new copy::

             Model(epc_file = 'new_file.epc', copy_from = 'existing.epc')
    """

    def __init__(self,
                 epc_file: Optional[str] = None,
                 full_load: bool = True,
                 epc_subdir: Optional[Union[str, Iterable]] = None,
                 new_epc: bool = False,
                 create_basics: Optional[bool] = None,
                 create_hdf5_ext: Optional[bool] = None,
                 copy_from: Optional[str] = None,
                 quiet = False):
        """Create an empty model; load it from epc_file if given.

        Note:

           if epc_file is given and the other arguments indicate that it will be a new dataset (new_epc is True
           or copy_from is given) then any existing .epc and .h5 file(s) with this name will be deleted
           immediately

        Arguments:
           epc_file (str, optional): if present, and new_epc is False and copy_from is None, the name
              of an existing epc file which is opened, unzipped and parsed to determine the list of parts
              and relationships comprising the model; if present, and new_epc is True or copy_from is
              specified, the name of a new file to be created - any existing file (and .h5 paired hdf5
              file) will be immediately deleted
              if None, an empty model is created (ie. with no parts) unless copy_from is present
           full_load (boolean): only relevant if epc_file is not None and new_epc is False (or copy_from
              is specified); if True (recommended), the xml for each part is parsed and stored in a tree
              structure in memory; if False, only the list of parts is loaded
           epc_subdir (string or list of strings, optional): if present, parts are only included in the load
              if they are in the top level directory of the epc internal structure, or in the specified
              subdirectory (or one of the subdirectories in the case of a list); only relevant if epc_file
              is not None and new_epc is False (or copy_from is specified)
           new_epc (boolean, default False): if True, a new model is created, empty unless copy_from is given
           create_basics (boolean, optional): if True and epc_file is None or new_epc is True,
              then the minimum essential parts are added to the empty Model; this is equivalent to
              calling the create_root(), create_rels_part() and create_doc_props() methods; if None, defaults
              to the same value as new_epc
           create_hdf5_ref (boolean, optional): if True and new_epc is True and create_basics is True
              and epc_file is not None, then an hdf5 external part is created, equivalent to calling the
              create_hdf5_ext() method; an empty hdf5 file is also created; if None, defaults to same
              value as new_epc
           copy_from: (str, optional): if present, and epc_file is also present, then the epc file
              named in copy_from, together with its paired h5 file, are copied to epc_file (overwriting
              any previous instances) before epc_file is opened; this argument is primarily to facilitate
              repeated testing of code that modifies the resqml dataset, eg. by appending new parts
           quiet (boolean, default False): if True, info logging message is emitted as debug

        Returns:
           The newly created Model object

        :meta common:
        """

        info_fn = log.debug if quiet else log.info

        if epc_file and not epc_file.endswith('.epc'):
            epc_file += '.epc'
        if copy_from and not copy_from.endswith('.epc'):
            copy_from += '.epc'
        if copy_from == epc_file:
            copy_from = None
        if create_basics is None:
            create_basics = new_epc and not copy_from
        if create_hdf5_ext is None:
            create_hdf5_ext = new_epc and not copy_from
        self.initialize()
        if epc_file and (copy_from or not new_epc):
            self.load_epc(epc_file,
                          full_load = full_load,
                          epc_subdir = epc_subdir,
                          copy_from = copy_from,
                          quiet = quiet)
        else:
            if epc_file and new_epc:
                try:
                    h5_file = epc_file[:-4] + '.h5'
                    os.remove(h5_file)
                    info_fn('old hdf5 file deleted: ' + str(h5_file))
                except Exception:
                    pass
                try:
                    os.remove(epc_file)
                    info_fn('old epc file deleted: ' + str(epc_file))
                except Exception:
                    pass
            if epc_file:
                self.set_epc_file_and_directory(epc_file)
            if create_basics:
                self.create_root()
                self.create_rels_part()
                self.create_doc_props()
                if epc_file and create_hdf5_ext:
                    assert epc_file.endswith('.epc')
                    h5_file = epc_file[:-4] + '.h5'
                    self.create_hdf5_ext(add_as_part = True, file_name = h5_file)
                    with h5py.File(h5_file, 'w') as _:
                        pass

    def initialize(self):
        """Set model contents to empty.

        note:
           not usually called directly (semi-private)
        """

        self.epc_file = None
        self.epc_directory = None
        # hdf5 stuff
        self.h5_dict = {}  # dictionary keyed on hdf5 uuid.bytes; mapping to hdf5 file name (full path)
        self.h5_currently_open_path = None
        self.h5_currently_open_root = None  # h5 file handle for open hdf5 file
        self.h5_currently_open_mode = None
        self.default_h5_override = 'full'  # one of 'none', 'dir', or 'full'
        self.main_h5_uuid = None  # uuid of main hdf5 file
        # xml stuff
        self.main_tree = None
        self.main_root = None
        self.crs_uuid = None  # primary coordinate reference system for model
        self.grid_root = None  # extracted from tree as speed optimization (useful for single grid models), for 'main' grid
        self.time_series = None  # extracted as speed optimization (single time series only for now)
        self.parts_forest = {}  # dictionary keyed on part_name; mapping to (content_type, uuid, xml_tree)
        self.uuid_part_dict = {}  # dictionary keyed on uuid.int; mapping to part_name
        self.rels_present = False
        self.rels_forest = {}  # dictionary keyed on part_name; mapping to (uuid, xml_tree)
        self.other_forest = {}  # dictionary keyed on part_name; mapping to (content_type, xml_tree); used for docProps
        # grid(s): single grid models only for now
        self.grid_list = []  # list of grid.Grid objects
        self.main_grid = None  # grid.Grid object for the 'main' grid
        self.reservoir_dict = [
        ]  # todo: mapping from reservoir name (citation title) to list of grids for that reservoir
        self.consolidation = None  # Consolidation object for mapping equivalent uuids
        self.modified = False

    def parts(self,
              parts_list = None,
              obj_type = None,
              uuid = None,
              title = None,
              title_mode = 'is',
              title_case_sensitive = False,
              metadata = {},
              extra = {},
              related_uuid = None,
              epc_subdir = None,
              sort_by = None):
        """Returns a list of parts matching all of the arguments passed.

        Arguments:
           parts_list (list of strings, optional): if present, an 'input' list of parts to be filtered;
              if None then all the parts in the model are considered
           obj_type (string, optional): if present, only parts of this resqml type will be included
           uuid (uuid.UUID, optional): if present, the uuid of a part to select
           title (string, optional): if present, a citation title or substring to filter on, based on
              the title_mode argument
           title_mode (string, default 'is'): one of 'is', 'starts', 'ends', 'contains',
              'is not', 'does not start', 'does not end', or 'does not contain'; how to compare each
              part's citation title with the title argument; ignored if title is None
           title_case_sensitive (boolean, default False): if True, title comparisons are made on a
              case sensitive basis; otherwise comparisons are insensitive to case
           metadata (dictionary of key:value pairs, optional): if present, only parts which have within
              their top level metadata all the items in this argument, are included in the filtered list
           extra (dictionary of key:value pairs, optional): if present, only parts which have within
              their extra metadata all the items in this argument, are included in the filtered list
           related_uuid (uuid.UUID, optional): if present, only parts which are related to this uuid
              are included in the filtered list
           epc_subdir (string, optional): if present, only parts which reside within the specified
              subdirectory path of the epc are included in the filtered list
           sort_by (string, optional): one of 'newest', 'oldest', 'title', 'uuid', 'type'

        Returns:
           a list of strings being the names of parts which match all filter arguments

        Examples:
           a full list of parts in the model::

              model.parts()

           a list of IjkGrid parts::

              model.parts(obj_type = 'IjkGridRepresentation')

           a list containing the part name for a uuid::

              model.parts(uuid = 'a869e7cc-5d30-4b31-8502-c74b1d87c777')

           a list of IjkGrid parts with titles beginning LGR, sorted by title::

              model.parts(obj_type = 'IjkGridRepresentation', title = 'LGR', title_mode = 'starts', sort_by = 'title')

        :meta common:
        """

        return m_c._parts(self,
                          parts_list = parts_list,
                          obj_type = obj_type,
                          uuid = uuid,
                          title = title,
                          title_mode = title_mode,
                          title_case_sensitive = title_case_sensitive,
                          metadata = metadata,
                          extra = extra,
                          related_uuid = related_uuid,
                          epc_subdir = epc_subdir,
                          sort_by = sort_by)

    def uuid_is_present(self, uuid):
        """Returns True if the uuid is present in the model's catalogue, False otherwise."""

        return m_f._uuid_is_present(self, uuid)

    def part(self,
             parts_list = None,
             obj_type = None,
             uuid = None,
             title = None,
             title_mode = 'is',
             title_case_sensitive = False,
             metadata = {},
             extra = {},
             related_uuid = None,
             epc_subdir = None,
             multiple_handling = 'exception'):
        """Returns the name of a part matching all of the arguments passed.

        arguments:
           (as for parts() except no sort_by argument)
           multiple_handling (string, default 'exception'): one of 'exception', 'none', 'first', 'oldest', 'newest'

        returns:
           string being the part name of the single part matching all of the criteria, or None

        notes:
           this method can be used where a single part is being identified; if no parts match the criteria, None is
           returned; if more than one part matches the criteria, the multiple_handling argument determines what happens:
           'exception' causes a ValueError exception to be raised; 'none' causes None to be returned; 'first' causes the
           first part (as stored in the epc file or added) to be returned; 'oldest' causes the part with the oldest
           creation timestamp in the citation block to be returned; 'newest' causes the newest part to be returned

        :meta common:
        """

        return m_c._part(self,
                         parts_list = parts_list,
                         obj_type = obj_type,
                         uuid = uuid,
                         title = title,
                         title_mode = title_mode,
                         title_case_sensitive = title_case_sensitive,
                         metadata = metadata,
                         extra = extra,
                         related_uuid = related_uuid,
                         epc_subdir = epc_subdir,
                         multiple_handling = multiple_handling)

    def uuids(self,
              parts_list = None,
              obj_type = None,
              uuid = None,
              title = None,
              title_mode = 'is',
              title_case_sensitive = False,
              metadata = {},
              extra = {},
              related_uuid = None,
              epc_subdir = None,
              sort_by = None):
        """Returns a list of uuids of parts matching all of the arguments passed.

        arguments:
           (as for parts() method)

        returns:
           list of uuids

        :meta common:
        """

        return m_c._uuids(self,
                          parts_list = parts_list,
                          obj_type = obj_type,
                          uuid = uuid,
                          title = title,
                          title_mode = title_mode,
                          title_case_sensitive = title_case_sensitive,
                          metadata = metadata,
                          extra = extra,
                          related_uuid = related_uuid,
                          epc_subdir = epc_subdir,
                          sort_by = sort_by)

    def uuid(self,
             parts_list = None,
             obj_type = None,
             uuid = None,
             title = None,
             title_mode = 'is',
             title_case_sensitive = False,
             metadata = {},
             extra = {},
             related_uuid = None,
             epc_subdir = None,
             multiple_handling = 'exception'):
        """Returns the uuid of a part matching all of the arguments passed.

        arguments:
           (as for part())

        returns:
           uuid of the single part matching all of the criteria, or None

        :meta common:
        """

        return m_c._uuid(self,
                         parts_list = parts_list,
                         obj_type = obj_type,
                         uuid = uuid,
                         title = title,
                         title_mode = title_mode,
                         title_case_sensitive = title_case_sensitive,
                         metadata = metadata,
                         extra = extra,
                         related_uuid = related_uuid,
                         epc_subdir = epc_subdir,
                         multiple_handling = multiple_handling)

    def roots(self,
              parts_list = None,
              obj_type = None,
              uuid = None,
              title = None,
              title_mode = 'is',
              title_case_sensitive = False,
              metadata = {},
              extra = {},
              related_uuid = None,
              epc_subdir = None,
              sort_by = None):
        """Returns a list of xml root nodes of parts matching all of the arguments passed.

        arguments:
           (as for parts() method)

        returns:
           list of lxml.etree.Element objects

        :meta common:
        """

        return m_c._roots(self,
                          parts_list = parts_list,
                          obj_type = obj_type,
                          uuid = uuid,
                          title = title,
                          title_mode = title_mode,
                          title_case_sensitive = title_case_sensitive,
                          metadata = metadata,
                          extra = extra,
                          related_uuid = related_uuid,
                          epc_subdir = epc_subdir,
                          sort_by = sort_by)

    def root(self,
             parts_list = None,
             obj_type = None,
             uuid = None,
             title = None,
             title_mode = 'is',
             title_case_sensitive = False,
             metadata = {},
             extra = {},
             related_uuid = None,
             epc_subdir = None,
             multiple_handling = 'exception'):
        """Returns the xml root node of a part matching all of the arguments passed.

        arguments:
           (as for part())

        returns:
           lxml.etree.Element object being the root node of the xml for the single part matching all of the criteria, or None

        :meta common:
        """

        return m_c._root(self,
                         parts_list = parts_list,
                         obj_type = obj_type,
                         uuid = uuid,
                         title = title,
                         title_mode = title_mode,
                         title_case_sensitive = title_case_sensitive,
                         metadata = metadata,
                         extra = extra,
                         related_uuid = related_uuid,
                         epc_subdir = epc_subdir,
                         multiple_handling = multiple_handling)

    def titles(self,
               parts_list = None,
               obj_type = None,
               uuid = None,
               title = None,
               title_mode = 'is',
               title_case_sensitive = False,
               metadata = {},
               extra = {},
               related_uuid = None,
               epc_subdir = None,
               sort_by = None):
        """Returns a list of citation titles of parts matching all of the arguments passed.

        arguments:
           (as for parts() method)

        returns:
           list of strings being the citation titles of matching parts

        :meta common:
        """

        return m_c._titles(self,
                           parts_list = parts_list,
                           obj_type = obj_type,
                           uuid = uuid,
                           title = title,
                           title_mode = title_mode,
                           title_case_sensitive = title_case_sensitive,
                           metadata = metadata,
                           extra = extra,
                           related_uuid = related_uuid,
                           epc_subdir = epc_subdir,
                           sort_by = sort_by)

    def title(self,
              parts_list = None,
              obj_type = None,
              uuid = None,
              title = None,
              title_mode = 'is',
              title_case_sensitive = False,
              metadata = {},
              extra = {},
              related_uuid = None,
              epc_subdir = None,
              multiple_handling = 'exception'):
        """Returns the citation title of a part matching all of the arguments passed.

        arguments:
           (as for part())

        returns:
           string being the citation title of the single part matching all of the criteria, or None

        :meta common:
        """

        return m_c._title(self,
                          parts_list = parts_list,
                          obj_type = obj_type,
                          uuid = uuid,
                          title = title,
                          title_mode = title_mode,
                          title_case_sensitive = title_case_sensitive,
                          metadata = metadata,
                          extra = extra,
                          related_uuid = related_uuid,
                          epc_subdir = epc_subdir,
                          multiple_handling = multiple_handling)

    def set_modified(self):
        """Marks the model as having been modified and assigns a new uuid.

        note:
           this modification tracking functionality is not part of the resqml standard and is only loosely
           applied by the library code; not usually called directly
        """

        self.modified = True

    @property
    def crs_root(self):
        """XML node corresponding to self.crs_uuid, the 'main' crs for the model."""

        return self.root_for_uuid(self.crs_uuid)

    def create_tree_if_none(self):
        """Checks that model has an xml tree; if not, an empty tree is created; not usually called directly."""

        m_x._create_tree_if_none(self)

    def load_part(self, epc, part_name, is_rels = None):
        """Load and parse xml tree for given part name, storing info in parts forest (or rels forest).

        arguments:
           epc: an open ZipFile handle for the epc file
           part_name (string): the name of the 'file' within the epc bundle containing the part (or relationship)
           is_rels: (boolean, optional): if True, the part to be loaded is a relationship part; if False,
              it is a main part; if None, its value is derived from the part name

        returns:
           boolean: True if part loaded successfully, False if part failed to load

        note:
           parts forest must already have been initialized before calling this method;
           if False is returned, calling code should probably delete part from forest;
           not usually called directly
        """

        return m_f._load_part(self, epc, part_name, is_rels = is_rels)

    def set_epc_file_and_directory(self, epc_file):
        """Sets the full path and directory of the epc_file.

        arguments:
           epc_file (string): the path of the epc file

        note:
           not usually needed to be called directly, except perhaps when creating a new dataset
        """

        self.epc_file = epc_file  # full path, if provided as such
        (self.epc_directory, _) = os.path.split(epc_file)
        if self.epc_directory is None or len(self.epc_directory) == 0:
            self.epc_directory = '.'

    def fell_part(self, part_name):
        """Removes the named part from the in-memory parts forest.

        arguments:
           part_name (string): the name of the part to be removed

        note:
           no check is made for references or relationships to the part being deleted;
           not usually called directly
        """

        m_f._fell_part(self, part_name)

    def remove_part_from_main_tree(self, part):
        """Removes the named part from the main (Content_Types) tree.

        note:
           not usually called directly
        """

        m_f._remove_part_from_main_tree(self, part)

    def tidy_up_forests(self, tidy_main_tree = True, tidy_others = False, remove_extended_core = True):
        """Removes any parts that do not have any related data in dictionaries.

        note:
           not usually called directly
        """

        m_f._tidy_up_forests(self,
                             tidy_main_tree = tidy_main_tree,
                             tidy_others = tidy_others,
                             remove_extended_core = remove_extended_core)

    def load_epc(self, epc_file, full_load = True, epc_subdir = None, copy_from = None, quiet = False):
        """Load xml parts of model from epc file (HDF5 arrays are not loaded).

        Arguments:
           epc_file (string): the path of the epc file
           full_load (boolean, default True): if True (recommended), the xml for each part is parsed
              and stored in a tree structure in memory; if False, only the list of parts is loaded
           epc_subdir (string or list of strings, optional): if present, only parts in the top
              level directory within the epc structure, or in the specified subdirectory(ies) are
              included in the load
           copy_from (string, optional): if present, the .epc and .h5 are copied from this source
              to epc_file (and paired .h5) prior to opening epc_file; any previous files named
              as epc_file will be overwritten
           quiet (boolean, default False): if True, info logging message is emitted as debug

        Returns:
           None

        Note:
           when copy_from is specified, the entire contents of the source dataset are copied,
           regardless of the epc_subdir setting which only affects the subsequent load into memory
        """

        m_f._load_epc(self,
                      epc_file,
                      full_load = full_load,
                      epc_subdir = epc_subdir,
                      copy_from = copy_from,
                      quiet = quiet)

    def store_epc(self,
                  epc_file = None,
                  main_xml_name = '[Content_Types].xml',
                  only_if_modified = False,
                  quiet = False):
        """Write xml parts of model to epc file (HDF5 arrays are not written here).

        Arguments:
           epc_file (string): the name of the output epc file to be written (any existing file will be
              overwritten)
           main_xml_name (string, do not pass): this argument should not be passed as the resqml standard
              requires the default value; (the argument exists in code because the resqml standard value
              is based on a slight misunderstanding of the opc standard, so could perhaps change in
              future versions of resqml)
           only_if_modified (boolean, default False): if True, the epc file is only written if the model
              is flagged as having been modified (at least one part added or removed)
           quiet (boolean, default False): if True, info logging is emitted at debug level

        Returns:
           None

        Note:
           the main tree, parts forest and rels forest must all be up to date before calling this method

        :meta common:
        """

        m_f._store_epc(self,
                       epc_file = epc_file,
                       main_xml_name = main_xml_name,
                       only_if_modified = only_if_modified,
                       quiet = quiet)

    def parts_list_of_type(self, type_of_interest = None, uuid = None):
        """Returns a list of part names for parts of type of interest, optionally matching a uuid.

        arguments:
           type_of_interest (string): the resqml object class of interest, in string form, eg. 'obj_IjkGridRepresentation'
           uuid (uuid.UUID object, optional): if present, only a part with this uuid is included in the list

        returns:
           a list of strings being the part names which match the arguments

        note:
           usually either a type of interest or a uuid is passed; if neither are passed, all parts are returned;
           this method is maintained for backward compatibility and for efficiency reasons;
           it is equivalent to: self.parts(obj_type = type_of_interest, uuid = uuid)
        """

        return m_c._parts_list_of_type(self, type_of_interest = type_of_interest, uuid = uuid)

    def list_of_parts(self, only_objects = True):
        """Return a complete list of parts."""

        return m_c._list_of_parts(self, only_objects = only_objects)

    def number_of_parts(self):
        """Retuns the number of parts in the model, including external parts such as the link to an hdf5 file."""

        return m_c._number_of_parts(self)

    def part_for_uuid(self, uuid):
        """Returns the part name which has the given uuid.

        arguments:
           uuid (uuid.UUID object or string): the uuid of the part of interest

        returns:
           a string being the part name which matches the uuid, or None if not found

        :meta common:
        """

        return m_c._part_for_uuid(self, uuid)

    def root_for_uuid(self, uuid):
        """Returns the xml root for the part which has the given uuid.

        arguments:
           uuid (uuid.UUID object or string): the uuid of the part of interest

        returns:
           the xml root node for the part with the given uuid, or None if not found

        :meta common:
        """

        return m_c._root_for_uuid(self, uuid)

    def parts_count_by_type(self, type_of_interest = None):
        """Returns a sorted list of (type, count) for parts.

        arguments:
           type_of_interest (string, optional): if not None, the returned list only contains one pair, with
              a count for that type (resqml object class)

        returns:
           list of pairs, each being (string, int) representing part type (resqml object class without leading obj_)
           and count
        """

        return m_c._parts_count_by_type(self, type_of_interest = type_of_interest)

    def parts_list_filtered_by_related_uuid(self, parts_list, uuid, uuid_is_source = None):
        """From a list of parts, returns a list of those parts which have a relationship with the given uuid.

        arguments:
           parts_list (list of strings): input list of parts from which a selection is made
           uuid (uuid.UUID): the uuid of a part for which related parts are required
           uuid_is_source (boolean, default None): if None, relationships in either direction qualify;
              if True, only those where uuid is sourceObject qualify; if False, only those where
              uuid is destinationObject qualify

        returns:
           list of strings being the subset of parts_list which are related to the object with the
           given uuid

        note:
           the part to which the given uuid applies might or might not be in the input parts list;
           this method scans the relationship info for every present part, looking for uuid in rels
        """

        return m_c._parts_list_filtered_by_related_uuid(self, parts_list, uuid, uuid_is_source = uuid_is_source)

    def supporting_representation_for_part(self, part):
        """Returns the uuid of the supporting representation for the part, if found, otherwise None."""

        return m_c._supporting_representation_for_part(self, part)

    def parts_list_filtered_by_supporting_uuid(self, parts_list, uuid):
        """From a list of parts, returns a list of those parts which have the given uuid as supporting representation.

        arguments:
           parts_list (list of strings): input list of parts from which a selection is made
           uuid (uuid.UUID): the uuid of a supporting representation part for which related parts are required

        returns:
           list of strings being the subset of parts_list which have as their supporting representation
           the object with the given uuid

        note:
           the part to which the given uuid applies might or might not be in the input parts list
        """

        return m_c._parts_list_filtered_by_supporting_uuid(self, parts_list, uuid)

    def parts_list_related_to_uuid_of_type(self, uuid, type_of_interest = None):
        """Returns a list of parts of type of interest that relate to part with given uuid.

        arguments:
           uuid (uuid.UUID): the uuid of a part for which related parts are required
           type_of_interest (string): the type of parts (resqml object class) of the related
              parts of interest

        returns:
           list of strings being the part names of the type of interest, related to the uuid
        """

        return m_c._parts_list_related_to_uuid_of_type(self, uuid, type_of_interest = type_of_interest)

    def external_parts_list(self):
        """Returns a list of part names for external part references.

        Returns:
           list of strings being the part names for external part references

        Note:

           in practice, external part references are only used for hdf5 files;
           furthermore, all current datasets have adopted the practice of using
           a single hdf5 file for a given epc file
        """

        return m_c._external_parts_list(self)

    def uuid_for_part(self, part_name, is_rels = None):
        """Returns the uuid for the named part.

        arguments:
           part_name (string): the part name for which the uuid is required
           is_rels (boolean, optional): if True, the part is a relationship part;
              if False, it is a main part; if None, its value is determined from the part name

        returns:
           uuid.UUID for the specified part

        note:
           this method will fail with an exception if the part is not in this model; a quicker alternative
           to this method is simply to extract the uuid from the part name using olio.xml_et.uuid_in_part_name()

        :meta common:
        """

        return m_c._uuid_for_part(self, part_name, is_rels = is_rels)

    def uuid_for_root(self, root_node):
        """Returns the uuid for an object given an xml root node.

        arguments:
           root_node: the xml root node for the object for which the uuid is required

        returns:
           uuid.UUID for the specified object
        """

        return rqet.uuid_for_part_root(root_node)

    def type_of_part(self, part_name, strip_obj = False):
        """Returns content type for the named part (does not apply to rels parts).

        arguments:
           part_name (string): the part for which the type is required
           strip_obj (boolean, default False): if True, the leading 'obj_' is removed
              from the returned string

        returns:
           string being the type (resqml object class) for the named part

        :meta common:
        """

        return m_c._type_of_part(self, part_name, strip_obj = strip_obj)

    def type_of_uuid(self, uuid, strip_obj = False):
        """Returns content type for the uuid.

        arguments:
           uuid (uuid.UUID or str): the uuid for which the type is required
           strip_obj (boolean, default False): if True, the leading 'obj_' is removed
              from the returned string

        returns:
           string being the type (resqml object class) for the named part

        :meta common:
        """

        return m_c._type_of_uuid(self, uuid, strip_obj = strip_obj)

    def tree_for_part(self, part_name, is_rels = None):
        """Returns parsed xml tree for the named part.

        arguments:
           part_name (string): the part name for which the xml tree is required
           is_rels (boolean, optional): if True, the part is a relationship part;
              if False, it is a main part; if None, its value is determined from the part name

        returns:
           parsed xml tree (defined in lxml or ElementTree package) for the named part
        """

        return m_c._tree_for_part(self, part_name, is_rels = is_rels)

    def root_for_part(self, part_name, is_rels = None):
        """Returns root of parsed xml tree for the named part.

        arguments:
           part_name (string): the part name for which the root of the xml tree is required
           is_rels (boolean, optional): if True, the part is a relationship part;
              if False, it is a main part; if None, its value is determined from the part name

        returns:
           root node of the parsed xml tree (defined in lxml or ElementTree package) for the named part

        :meta common:
        """

        return m_c._root_for_part(self, part_name, is_rels = is_rels)

    def change_hdf5_uuid_in_hdf5_references(self, node, old_uuid, new_uuid):
        """Scan node for hdf5 references and set the uuid of the hdf5 file itself to new_uuid.

        arguments:
           node: the root node of an xml tree within which hdf5 internal paths are to have hdf5 uuids changed
           old_uuid (uuid.UUID or str): the ext uuid currently to be found in the hdf5 references; if None, all will be replaced
           new_uuid (uuid.UUID or str): the new ext (hdf5) uuid to replace the old one

        returns:
           None

        note:
           use this method when the uuid of the hdf5 ext part is changing; if the uuid of the high level part itself is changing
           use change_uuid_in_hdf5_references() instead
        """

        m_h._change_hdf5_uuid_in_hdf5_references(self, node, old_uuid, new_uuid)

    def change_uuid_in_hdf5_references(self, node, old_uuid, new_uuid):
        """Scan node for hdf5 references using the old_uuid and replace with the new_uuid.

        arguments:
           node: the root node of an xml tree within which hdf5 internal paths are to have uuids changed
           old_uuid (uuid.UUID or str): the uuid currently to be found in the hdf5 references
           new_uuid (uuid.UUID or str): the new uuid to replace the old one

        returns:
           None

        notes:
           use this method when the uuid of the high level part itself is changing; if the uuid of the hdf5 ext part
           itself is changing, use change_hdf5_uuid_in_hdf5_references() instead; this method does not modify the
           internal path names in the hdf5 file itself, if that has already been written
        """

        m_h._change_uuid_in_hdf5_references(self, node, old_uuid, new_uuid)

    def change_uuid_in_supporting_representation_reference(self, node, old_uuid, new_uuid, new_title = None):
        """Look for supporting representation reference using the old_uuid and replace with the new_uuid.

        arguments:
           node: the root node of an xml tree within which the supporting representation uuid is to be changed
           old_uuid (uuid.UUID or str): the uuid currently to be found in the supporting representation reference
           new_uuid (uuid.UUID or str): the new uuid to replace the old one
           new_title (string, optional): if present, the title stored in the xml reference block is changed to this

        returns:
           boolean: True if the change was carried out; False otherwise

        notes:
           this method is typically used to temporarily set a supporting representation to a locally mocked
           representation object when the actual supporting representation is not present in the dataset
        """

        return m_x._change_uuid_in_supporting_representation_reference(self,
                                                                       node,
                                                                       old_uuid,
                                                                       new_uuid,
                                                                       new_title = new_title)

    def change_filename_in_hdf5_rels(self, new_hdf5_filename = None):
        """Scan relationships forest for hdf5 external parts and patch in a new filename.

        arguments:
           new_hdf5_filename: the new filename to patch into the xml; if None, the epc filename is used with
              no directory path and with the extension changed to .h5

        returns:
           None

        notes:
           no check is made that the new filename is for an existing file;
           all hdf5 file references will be modified
        """

        m_h._change_filename_in_hdf5_rels(self, new_hdf5_filename = new_hdf5_filename)

    def copy_part(self, existing_uuid, new_uuid, change_hdf5_refs = False):
        """Makes a new part as a copy of an existing part with only a new uuid set; the new part can then be modified.

        arguments:
           existing_uuid (uuid.UUID): the uuid of the existing part
           new_uuid (uuid.UUID): the uuid to inject into the new part after copying of the xml tree
           change_hdf5_refs (boolean): if True, the new tree is scanned for hdf5 refs using the existing_uuid and
              they are replaced with the new_uuid

        returns:
           string being the new part name

        notes:
           Resqml objects have a unique identifier and should be considered immutable; therefore to modify an object,
           it should first be duplicated; this function does some of the xml work needed for such duplication: the
           xml tree is copied; the uuid attribute in the root node is changed; the new part is added to the parts
           forest (with its name matched to the new uuid);
           NB: relationships are not currently copied or modified;
           also note that hdf5 data and high level objects maintained by other modules are not duplicated here;
           use this method to duplicate a part within a model prior to modifying the duplicated part in some way;
           to import a part from another model, use copy_part_from_other_model() instead; for copying a grid
           it is best to use the higher level derived_model.copy_grid() function
        """

        return m_f._copy_part(self, existing_uuid, new_uuid, change_hdf5_refs = change_hdf5_refs)

    def root_for_ijk_grid(self, uuid = None, title = None):
        """Return root for IJK Grid part.

        arguments:
           uuid (uuid.UUID, optional): if present, the uuid of the ijk grid part for which the root is required;
              if None, a single ijk grid part is expected and the root for that part is returned
           title (string, optional): if present, the citation title for the grid; defaults to 'ROOT' if more
              than one ijk grid present and no uuid supplied

        returns:
           root node in xml tree for the ijk grid part in this model

        notes:
           if uuid and title are both supplied, they must match in the corresponding grid part;
           if a title but no uuid is given, the first ijk grid encountered that has a matching title will be returned;
           if neither title nor uuid are given, the first ijk grid with title 'ROOT' will be returned, unless there is
           only one grid part in which case the root npde for that part is returned regardless;
           failure to find a matching grid part results in an assertion exception
        """

        return m_g._root_for_ijk_grid(self, uuid = uuid, title = title)

    def citation_title_for_part(self, part):  # duplicate functionality to title_for_part()
        """Returns the citation title for the specified part.

        :meta common:
        """

        return m_c._citation_title_for_part(self, part)

    def root_for_time_series(self, uuid = None):
        """Return root for time series part.

        argument:
           uuid (uuid.UUID, optional): if present, the uuid of the time series part for which the root is required;
              if None, a single time series part is expected and the root for that part is returned

        returns:
           root node in xml tree for the time series part in this model

        note:
           if no uuid is given and the model contains more than one time series, the one with the earliest creation
           date is returned
        """

        return m_c._root_for_time_series(self, uuid = uuid)

    def resolve_grid_root(self, grid_root = None, uuid = None):
        """If grid root argument is None, returns the root for the IJK Grid part instead.

        arguments:
           grid_root (optional): if not None, this method simply returns this argument
           uuid (uuid.UUID, optional): if present, the uuid of the ijk grid part for which the root is required;
              if None, an ijk grid part is sought and the root for that part is returned

        returns:
           root node in xml tree for the ijk grid part in this model ('ROOT' grid if more than one present)

        notes:
           if grid_root and uuid are both None and there are multiple grids in the model, the oldest
           grid with a citation title of 'ROOT' will be returned; an exception will be raised if no
           grid part is present in the model
        """

        return m_g._resolve_grid_root(self, grid_root = grid_root, uuid = uuid)

    def grid(self, title = None, uuid = None, find_properties = True):
        """Returns a shared Grid (or RegularGrid) object for this model, by default the 'main' grid.

        arguments:
           title (string, optional): if present, the citation title of the IjkGridRepresentation
           uuid (uuid.UUID, optional): if present, the uuid of the IjkGridRepresentation
           find_properties (boolean, default True): if True, the property_collection attribute
              of the returned grid object will be populated

        returns:
           grid.Grid object for the specified grid or the main ijk grid part in this model

        note:
           if neither title nor uuid are given, the model should contain just one grid, or a grid named 'ROOT';
           unlike most classes of object, a central list of resqpy Grid objects can be maintained within
           a Model by using this method which will return a shared object from this list, instantiating a new
           object and adding it to the list when necessary; an assertion exception will be raised if a
           suitable grid part is not present in the model

        :meta common:
        """

        return m_g._grid(self, title = title, uuid = uuid, find_properties = find_properties)

    def add_grid(self, grid_object, check_for_duplicates = False):
        """Add grid object to list of shareable grids for this model.

        arguments:
           grid_object (grid.Grid object): the ijk grid object to be added to the list
              of grids in the model
           check_for_duplicates (boolean, default False): if True, a check is made for any
              grid objects already in the grid list with the same root as the new grid object

        returns:
           None
        """

        return m_g._add_grid(self, grid_object = grid_object, check_for_duplicates = check_for_duplicates)

    def grid_list_uuid_list(self):
        """Returns list of uuid's for the grid objects in the cached grid list."""

        return m_g._grid_list_uuid_list(self)

    def grid_for_uuid_from_grid_list(self, uuid):
        """Returns the cached grid object matching the given uuid, if found in the grid list, otherwise None."""

        return m_g._grid_for_uuid_from_grid_list(self, uuid)

    def resolve_time_series_root(self, time_series_root = None):
        """If time_series_root is None, finds the root for a time series in the model.

        arguments:
           time_series_root (optional): if not None, this method simply returns this argument

        returns:
           root node in xml tree for the time series part in this model, or None if there is
           no time series part

        note:
           an assertion exception will be raised if time_series_root is None and there
           is more than one time series part in the model
        """

        return m_c._resolve_time_series_root(self, time_series_root = time_series_root)

    def h5_set_default_override(self, override):
        """Sets the default hdf5 filename override mode for the model.

        arguments:
           override (str): 'none', 'dir' or 'full' being the override mode to use by default

        note:
           this mode will be used by default when determining a filename for accessing hdf5 data;
           see h5_file_name() notes for more information
        """

        m_h._h5_set_default_override(self, override)

    def h5_uuid_and_path_for_node(self, node, tag = 'Values'):
        """Returns a (hdf5_uuid, hdf5_internal_path) pair for an xml array node.

        arguments:
           node: xml node for which the array reference is required
           tag (string, default 'Values'): the tag of the child of node for which the
              array reference is required

        returns:
           (uuid.UUID, string) pair being the uuid of the hdf5 external part reference and
           the hdf5 internal path for the array of interest

        note:
           this method provides the key data needed to actually access array data within
           the resqml dataset
        """

        return m_h._h5_uuid_and_path_for_node(self, node, tag = tag)

    def h5_uuid_list(self, node):
        """Returns a list of all uuids for hdf5 external part(s) referred to in recursive tree."""

        return m_h._h5_uuid_list(self, node)

    def h5_uuid(self):
        """Returns the uuid of the 'main' hdf5 file."""

        return m_h._h5_uuid(self)

    def h5_file_name(self, uuid = None, override = 'default', file_must_exist = True):
        """Returns full path for hdf5 file with given uuid.

        arguments:
           uuid (uuid.UUID, optional): the uuid of the hdf5 external part reference for which the
              file name is required; if None, the 'main' hdf5 uuid is used
           override (str or bool, default 'default'): if str, one of 'default', 'none', 'dir' or 'full';
              if bool (deprecated), False means 'dir' and True means 'full';
              if 'default', the default h5 override mode for the model is used;
              if 'dir', any directory in the rels hdf5 file name is replaced with the epc's directory;
              if 'full', the hdf5 full path is generated by using the epc path but replacing the .epc
              extension with .h5
           file_must_exist (boolean, default True): if True, the existence of the hdf5 file is checked
           upon first call and a FileNotFound exception is raised if the file is not found

        returns:
           string being the full path of the hdf5 file

        notes:
           in practice, a resqml model usually consists of a pair of files in the same directory,
           with names like: a.epc and a.h5;
           to allow copying, moving and renaming of files, the practical approach is simply
           to assume a one-to-one correspondence between epc and hdf5 files, and assume they
           are in the same directory, which will be the default behaviour of resqpy;
           to change the default behaviour for the model, call the h5_set_default_override() method;
           an hdf5 file name is cached once determined for a given ext uuid; to clear the cache,
           call the h5_clear_filename_cache() method
        """

        return m_h._h5_file_name(self, uuid = uuid, override = override, file_must_exist = file_must_exist)

    def h5_access(self, uuid = None, mode = 'r', override = 'default', file_path = None):
        """Returns an open h5 file handle for the hdf5 file with the given uuid.

        arguments:
           uuid (uuid.UUID): the uuid of the hdf5 external part reference for which the
              open file handle is required; required if override is False and file_path is None
           mode (string): the hdf5 file mode ('r', 'w' or 'a') with which to open the file
           override (str or bool, default 'default'): if str, one of 'default', 'none', 'dir' or 'full';
              if bool (deprecated), False means 'dir' and True means 'full';
              if 'default', the default h5 override mode for the model is used;
              if 'dir', any directory in the rels hdf5 file name is replaced with the epc's directory;
              if 'full', the hdf5 full path is generated by using the epc path but replacing the .epc
              extension with .h5
           file_path (string, optional): if present, is used as the hdf5 file path, otherwise
              the path will be determined based on the uuid and override arguments

        returns:
           a file handle to the opened hdf5 file

        note:
           an exception will be raised if the hdf5 file cannot be opened; note that sometimes another
           piece of code accessing the file might cause a 'resource unavailable' exception
        """

        return m_h._h5_access(self, uuid = uuid, mode = mode, override = override, file_path = file_path)

    def h5_release(self):
        """Releases (closes) the currently open hdf5 file.

        returns:
           None

        :meta common:
        """

        m_h._h5_release(self)

    def h5_array_shape_and_type(self, h5_key_pair):
        """Returns the shape and dtype of the array, as stored in the hdf5 file.

        arguments:
           h5_key_pair (uuid.UUID, string): the uuid of the hdf5 external part reference and the hdf5 internal path for the array

        returns:
           (tuple of ints, type): simply the shape and dtype attributes of the referenced hdf5 array; (None, None) is returned
           if the hdf5 file is not found, or the array is not found within it
        """

        return m_h._h5_array_shape_and_type(self, h5_key_pair)

    def h5_array_element(self,
                         h5_key_pair,
                         index = None,
                         cache_array = False,
                         object = None,
                         array_attribute = None,
                         dtype = 'float',
                         required_shape = None):
        """Returns one element from an hdf5 array and/or caches the array.

        arguments:
           h5_key_pair (uuid.UUID, string): the uuid of the hdf5 external part reference and the hdf5 internal path for the array
           index (pair or triple int, optional): if None, the only purpose of the call is to ensure that the array is cached in
              memory; if not None, the (k0, pillar_index) or (k0, j0, i0) index of the cell for which the value is required
           cache_array (boolean, default False): if True, a copy of the whole array is cached in memory as an attribute of the
              object; if already cached, the array is not uncached, regardless of this argument
           object (optional, defaults to self): the object in which a cached version of the array is an attribute, or will be
              created as an attribute if cache_array is True
           array_attribute (string): the attribute name to use for the cached version of the array,
              required to cache or access cached array
           dtype (string or data type): the data type of the elements of the array (need not match hdf5 array in precision)
           required_shape (tuple of ints, optional): if not None, the hdf5 array will be reshaped to this shape; if index
              is not None, it is taken to be applicable to the required shape

        returns:
           if index is None, then None;
           if index is not None, then the value of the array for the cell identified by index

        note:
           this function can be used to access an individual element from an hdf5 array, or to cache a whole array in memory;
           when accessing an individual element, the index style must match the array indexing; in particular for IJK grid points,
           a (k0, pillar_index) is needed when the grid has split pillars, whereas a (k0, j0, i0) is needed when the grid does
           not have any split pillars
        """

        return m_h._h5_array_element(self,
                                     h5_key_pair,
                                     index = index,
                                     cache_array = cache_array,
                                     object = object,
                                     array_attribute = array_attribute,
                                     dtype = dtype,
                                     required_shape = required_shape)

    def h5_array_slice(self, h5_key_pair, slice_tuple):
        """Loads a slice of an hdf5 array.

        arguments:
           h5_key_pair (uuid, string): the uuid of the hdf5 ext part and the hdf5 internal path to the
              required hdf5 array
           slice_tuple (tuple of slice objects): each element should be constructed using the python built-in
              function slice()

        returns:
           numpy array that is a hyper-slice of the hdf5 array, with the same ndim as the source hdf5 array

        notes:
           this method always fetches from the hdf5 file and does not attempt local caching; the whole array
           is not loaded; all axes continue to exist in the returned array, even where the sliced extent of
           an axis is 1
        """

        return m_h._h5_array_slice(self, h5_key_pair, slice_tuple)

    def h5_overwrite_array_slice(self, h5_key_pair, slice_tuple, array_slice):
        """Overwrites (updates) a slice of an hdf5 array.

        arguments:
           h5_key_pair (uuid, string): the uuid of the hdf5 ext part and the hdf5 internal path to the
              required hdf5 array
           slice_tuple (tuple of slice objects): each element should be constructed using the python built-in
              function slice()
           array_slice (numpy array of shape to match slice_tuple): the data to write

        notes:
           this method naively updates a slice in an hdf5 array without using mpi to look after parallel updates;
           metadata (such as uuid or property min, max values) is not modified in any way by the method
        """

        return m_h._h5_overwrite_array_slice(self, h5_key_pair, slice_tuple, array_slice)

    def h5_clear_filename_cache(self):
        """Clears the cached filenames associated with all ext uuids."""

        m_h._h5_clear_filename_cache(self)

    def create_root(self):
        """Initialises an empty main xml tree for model.

        note:
           not usually called directly
        """

        m_x._create_root(self)

    def add_part(self, content_type, uuid, root, add_relationship_part = True, epc_subdir = None):
        """Adds a (recently created) node as a new part in the model's parts forest.

        arguments:
           content_type (string): the resqml object class of the new part
           uuid (uuid.UUID): the uuid for the new part
           root: the root node of the xml tree for the new part
           add_relationship_part: (boolean, default True): if True, a relationship part is also
              created to go with the new part (empty of actual relationships)
           epc_subdir (string, optional): if present, the subdirectory path within the epc that the
              part is to be located within; if None, the xml will reside at the top level of the epc

        returns:
           None

        notes:
           NB: xml tree for part is not written to epc file by this function (store_epc() handles that);
           do not use this function for the main rels extension part: use create_rels_part() instead
        """

        m_f._add_part(self,
                      content_type,
                      uuid,
                      root,
                      add_relationship_part = add_relationship_part,
                      epc_subdir = epc_subdir)

    def patch_root_for_part(self, part, root):
        """Updates the xml tree for the part without changing the uuid."""

        m_f._patch_root_for_part(self, part, root)

    def remove_part(self, part_name, remove_relationship_part = True):
        """Removes a part from the parts forest; optionally remove corresponding rels part and other relationships."""

        m_f._remove_part(self, part_name, remove_relationship_part)

    def new_obj_node(self, flavour, name_space = 'resqml2', is_top_lvl_obj = True):
        """Creates a new main object element and sets attributes (does not add children).

        arguments:
           flavour (string): the resqml object class (type of part) for which a new xml tree
              is required
           name_space (string, default 'resqml2'): the xml namespace identifier to use for the node
           is_top_lvl_obj (boolean, default True): if True, the xsi:type is set in the xml node,
              as required for top level objects (parts); if False, the type atttribute is not set

        returns:
           newly created root node for xml tree for flavour of object, without any children
        """

        return m_x._new_obj_node(flavour, name_space = name_space, is_top_lvl_obj = is_top_lvl_obj)

    def referenced_node(self, ref_node, consolidate = False):
        """For a given xml reference node, returns the node for the object referred to, if present."""

        # note: the RESQML standard allows referenced objects to be missing from the package (model)

        return m_x._referenced_node(self, ref_node, consolidate = consolidate)

    def create_ref_node(self, flavour, title, uuid, content_type = None, root = None):
        """Create a reference node, optionally add to root.

        arguments:
           flavour (string): the resqml object class (type of part) for which a
              reference node is required
           title (string): used as the Title subelement text in the reference node
           uuid: (uuid.UUID): the uuid of the part being referenced
           content_type (string, optional): if None, the referenced content type is
              determined from the flavour argument (recommended)
           root (optional): if not None, an xml node to which the reference node is
              appended as a child

        returns:
           newly created reference xml node
        """

        return m_x._create_ref_node(flavour, title, uuid, content_type = content_type, root = root)

    def uom_node(self, root, uom):
        """Add a generic unit of measure sub element to root.

        arguments:
           root: xml node to which unit of measure subelement (child) will be added
           uom (string): the resqml unit of measure

        returns:
           newly created unit of measure node (having already been added to root)

        note:
           does not currently check that uom is a valid Energistics unit of measure;
           use weights_and_measures module functionality to check if needed
        """

        return m_x._uom_node(root, uom)

    def create_rels_part(self):
        """Adds a relationships reference node as a new part in the model's parts forest.

        returns:
           newly created main relationships reference xml node

        note:
           there can only be one relationships reference part in the model
        """

        return m_x._create_rels_part(self)

    def create_citation(self, root = None, title = '', originator = None):
        """Creates a citation xml node and optionally appends as a child of root.

        arguments:
           root (optional): if not None, the newly created citation node is appended as
              a child to this node
           title (string): the citation title: a human readable string; this is the main point
              of having a citation node, so the argument should be used wisely
           originator (string, optional): the name of the human being who created the object
              which this citation is for; default is to use the login name

        returns:
           newly created citation xml node
        """

        return m_x._create_citation(root = root, title = title, originator = originator)

    def title_for_root(self, root = None):
        """Returns the Title text from the Citation within the given root node.

        arguments:
           root: the xml node for the object for which the citation title is required

        returns:
           string being the Title text from the citation node which is a child of root,
           or None if not found

        :meta common:
        """

        return m_c._title_for_root(self, root = root)

    def title_for_part(self, part_name):  # duplicate functionality to citation_title_for_part()
        """Returns the Title text from the Citation for the given main part name (not for rels).

        arguments:
           part_name (string): the name of the part for which the citation title is required

        returns:
           string being the Title text from the citation node which is a child of the root xml node
           for the part, or None if not found

        :meta common:
        """

        return m_c._title_for_part(self, part_name)

    def create_unknown(self, root = None):
        """Creates an Unknown node and optionally adds as child of root.

        arguments:
           root (optional): if present, the newly created Unknown node is appended as a child
              of this xml node

        returns:
           the newly created Unknown xml node

        :meta common:
        """

        return m_x._create_unknown(root = root)

    def create_doc_props(self, add_as_part = True, root = None, originator = None):
        """Creates a document properties stub node and optionally adds as child of root and/or to parts forest.

        arguments:
           add_as_part (boolean, default True): if True, the newly created node is also added as a part
           root (optional, usually None): if not None, the newly created node is appended to this root
              as a child
           originator (string, optional): used as the creator in the doc props node; if None, the
              login name is used

        returns:
           the newly created doc props xml node

        note:
           the doc props part of a resqml dataset is intended to hold documentation and
           other stuff that is not covered by the standard; there should be exactly one doc props part
        """

        return m_x._create_doc_props(self, add_as_part = add_as_part, root = root, originator = originator)

    def create_crs_reference(self, root = None, crs_uuid = None):
        """Creates a node refering to an existing crs node and optionally adds as child of root.

        arguments:
           root: the xml node to which the new reference node is to appended as a child (ie. the xml node
              for the object that is referring to the crs)
           crs_uuid: the uuid of the crs

        returns:
           newly created crs reference xml node
        """

        return m_x._create_crs_reference(self, root = root, crs_uuid = crs_uuid)

    def create_md_datum_reference(self, md_datum_root, root = None):
        """Creates a node refering to an existing measured depth datum and optionally adds as child of root.

        arguments:
           md_datum_root: the root xml node for the measured depth datum being referenced
           root: the xml node to which the new reference node is to appended as a child (ie. the xml node
              for the object that is referring to the md datum)

        returns:
           newly created measured depth datum reference xml node
        """

        return m_x._create_md_datum_reference(md_datum_root, root = root)

    def create_hdf5_ext(self,
                        add_as_part = True,
                        root = None,
                        title = 'Hdf Proxy',
                        originator = None,
                        file_name = None,
                        uuid = None):
        """Creates an hdf5 external node and optionally adds as child of root and/or to parts forest.

        arguments:
           add_as_part (boolean, default True): if True the newly created ext node is added to the model
              as a part
           root (optional, usually None): if not None, the newly created ext node is appended as a child
              of this node
           title (string): used as the Title text in the citation node, usually left at the default 'Hdf Proxy'
           originator (string, optional): the name of the human being who created the ext object;
              default is to use the login name
           file_name (string, optional): the filename to be stored as the Target in the relationship node;
              if None, will default to the epc filename with the extenstion replaced with .h5
           uuid (uuid.UUID, optional): the ext uuid tp associate with the external part; if None, a new
              uuid will be generated

        returns:
           newly created hdf5 external part xml node

        note:
           this method is typically called when creating a new dataset (Model); if the intention is to
           share an existing hdf5 file, then pass the file_name and (ext) uuid arguments; if the intention
           is to create a new hdf5 file amongst many used by the Model, then pass the file_name
        """

        return m_x._create_hdf5_ext(self,
                                    add_as_part = add_as_part,
                                    root = root,
                                    title = title,
                                    originator = originator,
                                    file_name = file_name,
                                    uuid = uuid)

    def create_hdf5_dataset_ref(self, hdf5_uuid, object_uuid, group_tail, root, title = 'Hdf Proxy'):
        """Creates a pair of nodes referencing an hdf5 dataset (array) and adds to root.

        arguments:
           hdf5_uuid (uuid.UUID): the uuid of the hdf5 external part being referenced
           object_uuid (uuid.UUID): the uuid of the high level object (part) which owns the hdf5 array
              being referenced
           group_tail (string): the tail of the hdf5 internal path, which is appended to the part
              name section of the internal path
           root: the xml node to which the newly created hdf5 reference is appended as a child
           title (string): used as the Title text in the citation node, usually left at the default 'Hdf Proxy'

        returns:
           the newly created xml node holding the hdf5 internal path
        """

        return m_x._create_hdf5_dataset_ref(hdf5_uuid, object_uuid, group_tail, root, title = title)

    def create_supporting_representation(self,
                                         grid_root = None,
                                         support_uuid = None,
                                         root = None,
                                         title = None,
                                         content_type = 'obj_IjkGridRepresentation'):
        """Craate a supporting representation reference node refering to an IjkGrid and optionally add to root.

        arguments:
           grid_root: the xml node of the grid (or other) object which is the supporting representation
              being referred to; could also be used for other classes of supporting object; this or
              support_uuid must be provided
           support_uuid: the uuid of the grid (or other supporting representation) being referred to;
              this or grid_root must be provided
           root: if not None, the newly created supporting representation node is appended as a child
              to this node
           title: the Title to use in the supporting representation node
           content_type: the resqml object class of the supporting representation being referenced;
              defaults to 'obj_IjkGridRepresentation'

        returns:
           newly created xml node for supporting representation reference

        notes:
           a property array needs a supporting representation which is the structure that the property
           values belong to; for example, a grid property array has the grid object as the supporting
           representation; one of grid_root or support_uuid should be passed when calling this method
        """

        return m_x._create_supporting_representation(self,
                                                     support_root = grid_root,
                                                     support_uuid = support_uuid,
                                                     root = root,
                                                     title = title,
                                                     content_type = content_type)

    def create_source(self, source, root = None):
        """Create an extra meta data node holding information on the source of the data, optionally add to root.

        arguments:
           source (string): text describing the source of an object
           root: if not None, the newly created extra metadata node is appended as a child of this node

        returns:
           the newly created extra metadata xml node
        """

        return m_x._create_source(source, root = root)

    def create_patch(self,
                     p_uuid,
                     ext_uuid = None,
                     root = None,
                     patch_index = 0,
                     hdf5_type = 'DoubleHdf5Array',
                     xsd_type = 'double',
                     null_value = None,
                     const_value = None,
                     const_count = None,
                     points = False):
        """Create a node for a patch of values, including ref to hdf5 data set, optionally add to root.

        arguments:
           p_uuid (uuid.UUID): the uuid of the object for which this patch is a component
           ext_uuid (uuid.UUID): the uuid of the hdf5 external part holding the array; required unless
              const_value and const_count are not None
           root: if not None, the newly created patch of values xml node is appended as a child
              to this node
           patch_index (int, default 0): the patch index number; patches must be numbered
              sequentially starting at 0
           hdf5_type (string, default 'DoubleHdf5Array'): the type of the hdf5 array; usually one of
              'DoubleHdf5Array', 'IntegerHdf5Array', or 'BooleanHdf5Array'; replaced with equivalent
              constant array type if const_value is not None
           xsd_type (string, default 'double'): the xsd simple type of each element of the array
           null_value: Used in a null value sub-node to specify what value in an array of discrete data represents null;
              if None, a value of -1 is used for signed integers, 2^32 - 1 for uints (even 64 bit uints!)
           const_value (float, int, or bool, optional): if not None, the patch is created as a constant array;
              const_count must also be present if const_value is not None
           const_count (int, optional): the number of elements in (size of) the constant array; required if
              const_value is not None, ignored otherwise
           points (bool, default False): if True, the created node will be for a patch of points,
              otherwise a patch of values

        returns:
           newly created xml node for the patch of values

        note:
           this function does not write the data to the hdf5; that should be done separately before
           calling this method;
           RESQML usually stores array data in the hdf5 file however constant arrays are flagged as
           such in the xml and no data is stored in the hdf5
        """

        return m_x._create_patch(p_uuid,
                                 ext_uuid = ext_uuid,
                                 root = root,
                                 patch_index = patch_index,
                                 hdf5_type = hdf5_type,
                                 xsd_type = xsd_type,
                                 null_value = null_value,
                                 const_value = const_value,
                                 const_count = const_count,
                                 points = points)

    def create_time_series_ref(self, time_series_uuid, root = None):
        """Create a reference node to a time series, optionally add to root.

        arguments:
           time_series_uuid (uuid.UUID): the uuid of the time series part being referenced
           root (optional): if present, the newly created time series reference xml node is added
              as a child to this node

        returns:
           the newly created time series reference xml node
        """

        return m_x._create_ref_node('TimeSeries', 'time series', time_series_uuid, root = root)

    def create_solitary_point3d(self, flavour, root, xyz):
        """Creates a subelement to root for a solitary point in 3D space.

        arguments:
           flavour (string): the object class (type) of the point node to be created
           root: the xml node to which the newly created solitary point node is appended
              as a child
           xyz (triple float): the x, y, z coordinates of the solitary point

        returns:
           the newly created xml node for the solitary point
        """

        return m_x._create_solitary_point3d(flavour, root, xyz)

    def create_reciprocal_relationship(self, node_a, rel_type_a, node_b, rel_type_b, avoid_duplicates = True):
        """Adds a node to each of a pair of trees in the rels forest, to represent a two-way relationship.

        arguments:
           node_a: one of the two xml nodes to be related
           rel_type_a (string): the Type (role) associated with node_a in the relationship;
              usually 'sourceObject' or 'destinationObject'
           node_b: the other xml node to be related
           rel_type_b (string): the Type (role) associated with node_b in the relationship
              usually 'sourceObject' or 'destinationObject' (opposite of rel_type_a)
           avoid_duplicates (boolean, default True): if True, xml for a relationship is not added
              where it already exists; if False, a duplicate will be created in this situation

        returns:
           None

        note:
           this method has the same effect as create_reciprocal_relationship_uuids() but takes xml root nodes
           rather than uuids as arguments
        """

        return m_x._create_reciprocal_relationship(self,
                                                   node_a,
                                                   rel_type_a,
                                                   node_b,
                                                   rel_type_b,
                                                   avoid_duplicates = avoid_duplicates)

    def create_reciprocal_relationship_uuids(self, uuid_a, rel_type_a, uuid_b, rel_type_b, avoid_duplicates = True):
        """Adds a node to each of a pair of trees in the rels forest, to represent a two-way relationship.

        arguments:
           uuid_a: uuid of one of the two parts to be related
           rel_type_a (string): the Type (role) associated with uuid_a in the relationship;
              usually 'sourceObject' or 'destinationObject'
           uuid_b: uuid of the other part to be related
           rel_type_b (string): the Type (role) associated with uuid_b in the relationship
              usually 'sourceObject' or 'destinationObject' (opposite of rel_type_a)
           avoid_duplicates (boolean, default True): if True, xml for a relationship is not added
              where it already exists; if False, a duplicate will be created in this situation

        returns:
           None

        note:
           this method has the same effect as create_reciprocal_relationship() but takes uuids rather than
           xml root nodes as arguments
        """

        node_a = self.root_for_uuid(uuid_a)
        node_b = self.root_for_uuid(uuid_b)
        assert node_a is not None and node_b is not None, 'part not present when creating relationship'
        return m_x._create_reciprocal_relationship(self,
                                                   node_a,
                                                   rel_type_a,
                                                   node_b,
                                                   rel_type_b,
                                                   avoid_duplicates = avoid_duplicates)

    def duplicate_node(self, existing_node, add_as_part = True):
        """Creates a deep copy of the xml node (typically from another model) and optionally adds as part.

        arguments:
           existing_node: the existing xml node, usually in another model, to be duplicated
           add_as_part (boolean, default True): if True, the newly created xml node is added as a part
              in this model

        returns:
           the newly created duplicate xml node

        notes:
           hdf5 data is not copied by this function and any reference to hdf5 arrays will be naively
           duplicated;
           the uuid of the part is not changed by this function, so if the source and target models are
           the same, add as part should be set False and calling code will need to assign a new uuid
           prior to adding as part;
           if add_as_part is True, the part is only added if the uuid does not already exist in this
           model
        """

        return m_f._duplicate_node(self, existing_node, add_as_part = add_as_part)

    def force_consolidation_uuid_equivalence(self, immigrant_uuid, resident_uuid):
        """Force immigrant object to be teated as equivalent to resident during consolidation."""

        m_f._force_consolidation_uuid_equivalence(self, immigrant_uuid, resident_uuid)

    def force_consolidation_equivalence_for_class_ignoring_extra_metadata(self, other_model, resqpy_class):
        """Force immigrant objects of type to be teated as equivalent where only extra metadata differs during consolidation.

        notes:
            this method should be called prior to calling copy_part_from_other_model() or
            copy_uuid_from_other_model() to override the more stringent equivalence checks which
            include extra metadata; the resqpy class must have an is_equivalent() method which
            supports the check_extra_metadata boolean argument;
            typically used for organisational classes such as features and interpretations;
            an exception is raised if more than one matching part already exists in this model, for a
            particular immigrant title (for the object type)
        """

        obj_type = resqpy_class.resqml_type
        resident_uuids = self.uuids(obj_type = obj_type)
        if len(resident_uuids) == 0:
            log.debug(f'no resident parts for class {resqpy_class}')
            return
        immigrant_uuids = other_model.uuids(obj_type = obj_type)
        if len(immigrant_uuids) == 0:
            log.debug(f'no immigrant parts for class {resqpy_class} object type {obj_type}')
            return
        resident_objects_dict = {}
        for resident_uuid in resident_uuids:
            resident_objects_dict[resident_uuid] = resqpy_class(self, uuid = resident_uuid)
        for immigrant_uuid in immigrant_uuids:
            immigrant_obj = resqpy_class(other_model, uuid = immigrant_uuid)
            log.debug(f'looking for equivalent for {obj_type} {immigrant_obj.title}')
            for resident_uuid in resident_uuids:
                if resident_objects_dict[resident_uuid].is_equivalent(immigrant_obj, check_extra_metadata = False):
                    log.debug(f'forcing equivalence between immigrant {immigrant_uuid} and resident {resident_uuid} ' +
                              f'of class {resqpy_class} object type {obj_type}')
                    m_f._force_consolidation_uuid_equivalence(self, immigrant_uuid, resident_uuid)
                    break

    def remove_extra_metadata(self, uuid):
        """Removes extra metadata from in memory xml for uuid.

        note:
            this method will not modify any resqpy objects already instantiated
        """

        rqet.cut_extra_metadata(self.root_for_uuid(uuid))

    def copy_part_from_other_model(self,
                                   other_model,
                                   part,
                                   realization = None,
                                   consolidate = True,
                                   force = False,
                                   cut_refs_to_uuids = None,
                                   cut_node_types = None,
                                   self_h5_file_name = None,
                                   h5_uuid = None,
                                   other_h5_file_name = None):
        """Fully copies part in from another model, with referenced parts, hdf5 data and relationships.

        arguments:
           other model (Model): the source model from which to copy a part
           part (string): the part name in the other model to copy into this model
           realization (int, optional): if present and the part is a property, the realization
              will be set to this value, instead of the value in use in the other model if any
           consolidate (boolean, default True): if True and an equivalent part already exists in
              this model, do not duplicate but instead note uuids as equivalent
           force (boolean, default False): if True, the part itself is copied without much checking
              and all references are required to be handled by an entry in the consolidation object
           cut_refs_to_uuids (list of UUIDs, optional): if present, then xml reference nodes
              referencing any of the listed uuids are cut out in the copy; use with caution
           cut_node_types (list of str, optional): if present, any child nodes of a type in the list
              will be cut out in the copy; use with caution
           self_h5_file_name (string, optional): h5 file name for this model; can be passed as
              an optimisation when calling method repeatedly
           h5_uuid (uuid, optional): UUID for this model's hdf5 external part; can be passed as
              an optimisation when calling method repeatedly
           other_h5_file_name (string, optional): h5 file name for other model; can be passed as
              an optimisation when calling method repeatedly

        returns:
           the part name of the part in this model, after copying; may differ from requested part if
           consolidate is True; None in the case of failure

        notes:
           if the part name already exists in this model, no action is taken;
           default hdf5 file used in this model and assumed in other_model
        """

        return m_f._copy_part_from_other_model(self,
                                               other_model,
                                               part,
                                               realization = realization,
                                               consolidate = consolidate,
                                               force = force,
                                               cut_refs_to_uuids = cut_refs_to_uuids,
                                               cut_node_types = cut_node_types,
                                               self_h5_file_name = self_h5_file_name,
                                               h5_uuid = h5_uuid,
                                               other_h5_file_name = other_h5_file_name)

    def copy_uuid_from_other_model(self,
                                   other_model,
                                   uuid,
                                   realization = None,
                                   consolidate = True,
                                   force = False,
                                   cut_refs_to_uuids = None,
                                   cut_node_types = None,
                                   self_h5_file_name = None,
                                   h5_uuid = None,
                                   other_h5_file_name = None):
        """Fully copies part for uuid in from another model, with referenced parts, hdf5 data and relationships.

        arguments:
           other model (Model): the source model from which to copy a part
           uuid (UUID): the uuid of the part in the other model to copy into this model
           realization (int, optional): if present and the part is a property, the realization
              will be set to this value, instead of the value in use in the other model if any
           consolidate (boolean, default True): if True and an equivalent part already exists in
              this model, do not duplicate but instead note uuids as equivalent
           force (boolean, default False): if True, the part itself is copied without much checking
              and all references are required to be handled by an entry in the consolidation object
           cut_refs_to_uuids (list of UUIDs, optional): if present, then xml reference nodes
              referencing any of the listed uuids are cut out in the copy; use with caution
           cut_node_types (list of str, optional): if present, any child nodes of a type in the list
              will be cut out in the copy; use with caution
           self_h5_file_name (string, optional): h5 file name for this model; can be passed as
              an optimisation when calling method repeatedly
           h5_uuid (uuid, optional): UUID for this model's hdf5 external part; can be passed as
              an optimisation when calling method repeatedly
           other_h5_file_name (string, optional): h5 file name for other model; can be passed as
              an optimisation when calling method repeatedly

        returns:
           the uuid of the part in this model, after copying; may differ from requested uuid if
           consolidate is True; None in the case of failure

        notes:
           if the part already exists in this model, no action is taken;
           default hdf5 file used in this model and assumed in other_model
        """

        part = other_model.part_for_uuid(uuid)
        if part is None:
            return None
        copied_part = m_f._copy_part_from_other_model(self,
                                                      other_model,
                                                      part,
                                                      realization = realization,
                                                      consolidate = consolidate,
                                                      force = force,
                                                      cut_refs_to_uuids = cut_refs_to_uuids,
                                                      cut_node_types = cut_node_types,
                                                      self_h5_file_name = self_h5_file_name,
                                                      h5_uuid = h5_uuid,
                                                      other_h5_file_name = other_h5_file_name)
        if copied_part is None:
            return None
        return self.uuid_for_part(copied_part)

    def copy_all_parts_from_other_model(self, other_model, realization = None, consolidate = True):
        """Fully copies parts in from another model, with referenced parts, hdf5 data and relationships.

        arguments:
           other model (Model): the source model from which to copy parts
           realization (int, optional): if present, the realization attribute of property parts
              will be set to this value, instead of the value in use in the other model if any
           consolidate (boolean, default True): if True, where equivalent part already exists in
              this model, do not duplicate but instead note uuids as equivalent, modifying
              references and relationships of other copied parts appropriately

        notes:
           part names already existing in this model are not duplicated;
           default hdf5 file used in this model and assumed in other_model
        """

        m_f._copy_all_parts_from_other_model(self, other_model, realization = realization, consolidate = consolidate)

    def iter_objs(self, cls):
        """Iterate over all available objects of given resqpy class within the model.

        Note:

           The resqpy class must expose a class attribute `resqml_type`, and must support
           being created with the signature: `obj = cls(model, uuid=uuid)`.

        Example use::

           for well in model.iter_objs(cls=resqpy.well.WellboreFeature):
              print(well.title, well.uuid)

        Args:
           cls: resqpy class to iterate

        Yields:
           list of instances of cls

        :meta common:
        """

        for obj in m_c._iter_objs(self, cls):
            yield obj

    def iter_grid_connection_sets(self):
        """Yields grid connection set objects, one for each gcs in this model."""

        for gcs in list(m_c._iter_grid_connection_sets(self)):
            yield gcs

    def iter_wellbore_interpretations(self):
        """Iterable of all WellboreInterpretations associated with the model.

        Yields:
           wellbore: instance of :class:`resqpy.organize.WellboreInterpretation`

        :meta common:
        """

        for wi in m_c._iter_wellbore_interpretations(self):
            yield wi

    def iter_trajectories(self):
        """Iterable of all trajectories associated with the model.

        Yields:
           trajectory: instance of :class:`resqpy.well.Trajectory`

        :meta common:
        """

        for wt in m_c._iter_trajectories(self):
            yield wt

    def iter_md_datums(self):
        """Iterable of all MdDatum objects associated with the model.

        Yields:
           md_datum: instance of :class:`resqpy.well.MdDatum`

        :meta common:
        """

        for mdd in m_c._iter_md_datums(self):
            yield mdd

    def iter_crs(self):
        """Iterable of all CRS objects associated with the model.

        Yields:
           crs: instance of :class:`resqpy.crs.CRS`

        :meta common:
        """

        for crs in m_c._iter_crs(self):
            yield crs

    def sort_parts_list_by_timestamp(self, parts_list):
        """Returns a copy of the parts list sorted by citation block creation date, with the newest first."""

        return m_c._sort_parts_list_by_timestamp(self, parts_list)

    def as_graph(self, uuids_subset = None):
        """Return representation of model as nodes and edges, suitable for plotting in a graph.

        Note:
           The graph can be most readily visualised with other packages such as
           NetworkX and HoloViews, which are not part of resqpy.

           For a guide to plotting graphs interactively, see:
           http://holoviews.org/user_guide/Network_Graphs.html

        Example::

           # Create the nodes and edges
           nodes, edges = model.as_graph()

           # Load into a NetworkX graph
           import networkx as nx
           g = nx.Graph()
           g.add_nodes_from(nodes.items())
           g.add_edges_from(edges)

           # Import holoviews
           import holoviews as hv
           from holoviews import opts
           hv.extension('bokeh')

           # Plot
           hv.Graph.from_networkx(g, nx.layout.spring_layout).opts(
              tools=['hover'], node_color='resqml_type', cmap='Category10'
           )

        Args:
           uuids_subset (iterable): If present, only consider uuids in this list.
              By default, use all uuids in the model.

        Returns:
           2-tuple of nodes and edges:

           - nodes: dict mapping uuid to attributes (e.g. citation title)
           - edges: set of unordered pairs of uuids, representing relationships

        :meta common:
        """

        return m_c._as_graph(self, uuids_subset = uuids_subset)


def new_model(epc_file, quiet = False) -> Model:
    """Returns a new, empty Model object with basics and hdf5 ext part set up."""

    return Model(epc_file = epc_file, new_epc = True, create_basics = True, create_hdf5_ext = True, quiet = quiet)



"""Submodule containing functions for creating xml for a property collection."""

version = '1st December 2021'

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

import math as maths
import numpy as np

import resqpy.property
import resqpy.time_series as rts
import resqpy.weights_and_measures as wam
import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.property._collection_get_attributes as pcga
import resqpy.property.property_common as rqp_c
from resqpy.olio.xml_namespaces import curly_namespace as ns


def _create_xml_add_as_part(collection, add_as_part, p_uuid, p_node, add_relationships, support_root,
                            property_kind_uuid, related_time_series_node, sl_root, discrete, string_lookup_uuid,
                            const_value, ext_uuid):
    if add_as_part:
        collection.model.add_part('obj_' + collection.d_or_c_text + 'Property', p_uuid, p_node)
        if add_relationships:
            _create_xml_add_relationships(collection, p_node, support_root, property_kind_uuid,
                                          related_time_series_node, sl_root, discrete, string_lookup_uuid, const_value,
                                          ext_uuid)


def _create_property_set_xml_add_as_part(collection, ps_node, ps_uuid, add_relationships, parent_set_ref_node,
                                         prop_node_list):
    collection.model.add_part('obj_PropertySet', ps_uuid, ps_node)
    if add_relationships:
        # todo: add relationship with time series if time set kind is not 'not a time set'?
        if collection.parent_set_root is not None:
            collection.model.create_reciprocal_relationship(ps_node, 'destinationObject', parent_set_ref_node,
                                                            'sourceObject')
        for prop_node in prop_node_list:
            collection.model.create_reciprocal_relationship(ps_node, 'destinationObject', prop_node, 'sourceObject')


def _create_xml_get_basics(collection, discrete, points, const_value, facet_type, null_value, support_uuid, ext_uuid):
    assert not discrete or not points
    assert not points or const_value is None
    assert not points or facet_type is None
    assert collection.model is not None

    if null_value is not None:
        collection.null_value = null_value

    if support_uuid is None:
        support_uuid = collection.support_uuid
    assert support_uuid is not None
    support_root = collection.model.root_for_uuid(support_uuid)
    # assert support_root is not None

    if ext_uuid is None:
        ext_uuid = collection.model.h5_uuid()

    return support_root, support_uuid, ext_uuid


def _create_xml_property_kind(collection, p_node, find_local_property_kinds, property_kind, uom, discrete,
                              property_kind_uuid):
    if property_kind == 'permeability rock':
        property_kind = 'rock permeability'
    p_kind_node = rqet.SubElement(p_node, ns['resqml2'] + 'PropertyKind')
    p_kind_node.text = rqet.null_xml_text
    if find_local_property_kinds and property_kind not in rqp_c.supported_property_kind_list:
        property_kind_uuid = pcga._get_property_kind_uuid(collection, property_kind_uuid, property_kind, uom, discrete)

    if property_kind_uuid is None:
        p_kind_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'StandardPropertyKind')  # todo: local prop kind ref
        kind_node = rqet.SubElement(p_kind_node, ns['resqml2'] + 'Kind')
        kind_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'ResqmlPropertyKind')
        kind_node.text = property_kind
    else:
        p_kind_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'LocalPropertyKind')  # todo: local prop kind ref
        collection.model.create_ref_node('LocalPropertyKind',
                                         property_kind,
                                         property_kind_uuid,
                                         content_type = 'obj_PropertyKind',
                                         root = p_kind_node)
    return property_kind_uuid


def _create_xml_patch_node(collection, p_node, points, const_value, indexable_element, direction, p_uuid, ext_uuid,
                           expand_const_arrays):
    # create patch node
    const_count = None
    if const_value is not None and not expand_const_arrays:
        s_shape = collection.supporting_shape(indexable_element = indexable_element, direction = direction)
        assert s_shape is not None
        const_count = np.product(np.array(s_shape, dtype = int))
    else:
        const_value = None
    _ = collection.model.create_patch(p_uuid,
                                      ext_uuid,
                                      root = p_node,
                                      hdf5_type = collection.hdf5_type,
                                      xsd_type = collection.xsd_type,
                                      null_value = collection.null_value,
                                      const_value = const_value,
                                      const_count = const_count,
                                      points = points)


def _create_xml_property_min_max(collection, property_array, const_value, discrete, add_min_max, p_node, min_value,
                                 max_value, categorical, null_value, points):
    if add_min_max and not categorical and not points:
        # todo: use active cell mask on numpy min and max operations; exclude null values on discrete min max
        min_value, max_value = pcga._get_property_array_min_max_value(collection, property_array, const_value, discrete,
                                                                      min_value, max_value, categorical, null_value)
        if min_value is not None:
            min_node = rqet.SubElement(p_node, ns['resqml2'] + 'MinimumValue')
            min_node.set(ns['xsi'] + 'type', ns['xsd'] + collection.xsd_type)
            if discrete:
                min_node.text = str(maths.floor(min_value))
            else:
                min_node.text = str(min_value)
        if max_value is not None:
            max_node = rqet.SubElement(p_node, ns['resqml2'] + 'MaximumValue')
            max_node.set(ns['xsi'] + 'type', ns['xsd'] + collection.xsd_type)
            if discrete:
                max_node.text = str(maths.ceil(max_value))
            else:
                max_node.text = str(max_value)


def _create_xml_lookup_node(collection, p_node, string_lookup_uuid):
    sl_root = None
    if string_lookup_uuid is not None:
        sl_root = collection.model.root_for_uuid(string_lookup_uuid)
        assert sl_root is not None, 'string table lookup is missing whilst importing categorical property'
        assert rqet.node_type(sl_root) == 'obj_StringTableLookup', 'referenced uuid is not for string table lookup'
        collection.model.create_ref_node('Lookup',
                                         collection.model.title_for_root(sl_root),
                                         string_lookup_uuid,
                                         content_type = 'obj_StringTableLookup',
                                         root = p_node)
    return sl_root


def _create_xml_uom_node(collection, p_node, uom, property_kind, min_value, max_value, facet_type, facet, title,
                         points):
    if points:
        return
    if not uom:
        uom = rqp_c.guess_uom(property_kind,
                              min_value,
                              max_value,
                              collection.support,
                              facet_type = facet_type,
                              facet = facet)
        if not uom:
            uom = 'Euc'  # todo: put RESQML base uom for quantity class here, instead of Euc
            log.warning(f'uom set to Euc for property {title} of kind {property_kind}')
    if uom in wam.valid_uoms():
        collection.model.uom_node(p_node, uom)
    else:
        collection.model.uom_node(p_node, 'Euc')
        rqet.create_metadata_xml(p_node, {'uom': uom})


def _create_xml_add_relationships(collection, p_node, support_root, property_kind_uuid, related_time_series_node,
                                  sl_root, discrete, string_lookup_uuid, const_value, ext_uuid):
    if support_root is not None:
        collection.model.create_reciprocal_relationship(p_node, 'destinationObject', support_root, 'sourceObject')
    if property_kind_uuid is not None:
        pk_node = collection.model.root_for_uuid(property_kind_uuid)
        if pk_node is not None:
            collection.model.create_reciprocal_relationship(p_node, 'destinationObject', pk_node, 'sourceObject')
    if related_time_series_node is not None:
        collection.model.create_reciprocal_relationship(p_node, 'destinationObject', related_time_series_node,
                                                        'sourceObject')
    if discrete and string_lookup_uuid is not None:
        collection.model.create_reciprocal_relationship(p_node, 'destinationObject', sl_root, 'sourceObject')

    if const_value is None:
        ext_node = collection.model.root_for_part(
            rqet.part_name_for_object('obj_EpcExternalPartReference', ext_uuid, prefixed = False))
        collection.model.create_reciprocal_relationship(p_node, 'mlToExternalPartProxy', ext_node,
                                                        'externalPartProxyToMl')


def _create_xml_add_basics_to_p_node(collection, p_node, title, originator, extra_metadata, source, count,
                                     indexable_element):
    collection.model.create_citation(root = p_node, title = title, originator = originator)
    rqet.create_metadata_xml(node = p_node, extra_metadata = extra_metadata)

    if source is not None and len(source) > 0:
        collection.model.create_source(source = source, root = p_node)

    count_node = rqet.SubElement(p_node, ns['resqml2'] + 'Count')
    count_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
    count_node.text = str(count)

    ie_node = rqet.SubElement(p_node, ns['resqml2'] + 'IndexableElement')
    ie_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'IndexableElements')
    ie_node.text = indexable_element


def _create_xml_realization_node(realization, p_node):
    if realization is not None and realization >= 0:
        ri_node = rqet.SubElement(p_node, ns['resqml2'] + 'RealizationIndex')
        ri_node.set(ns['xsi'] + 'type', ns['xsd'] + 'nonNegativeInteger')
        ri_node.text = str(realization)


def _create_xml_time_series_node(collection, time_series_uuid, time_index, p_node, support_uuid, support_type,
                                 support_root):
    if time_series_uuid is None or time_index is None:
        related_time_series_node = None
    else:
        related_time_series_node = collection.model.root(uuid = time_series_uuid)
        time_series = rts.any_time_series(collection.model, uuid = time_series_uuid)
        time_series.create_time_index(time_index, root = p_node)

    support_title = '' if support_root is None else rqet.citation_title_for_node(support_root)
    collection.model.create_supporting_representation(support_uuid = support_uuid,
                                                      root = p_node,
                                                      title = support_title,
                                                      content_type = support_type)
    return related_time_series_node


def _create_xml_get_p_node(collection, p_uuid):
    p_node = collection.model.new_obj_node(collection.d_or_c_text + 'Property')
    if p_uuid is None:
        p_uuid = bu.uuid_from_string(p_node.attrib['uuid'])
    else:
        p_node.attrib['uuid'] = str(p_uuid)
    return p_node, p_uuid


def _create_xml_facet_node(facet_type, facet, p_node):
    if facet_type is not None and facet is not None:
        facet_node = rqet.SubElement(p_node, ns['resqml2'] + 'Facet')
        facet_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'PropertyKindFacet')
        facet_node.text = rqet.null_xml_text
        facet_type_node = rqet.SubElement(facet_node, ns['resqml2'] + 'Facet')
        facet_type_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Facet')
        facet_type_node.text = facet_type
        facet_value_node = rqet.SubElement(facet_node, ns['resqml2'] + 'Value')
        facet_value_node.set(ns['xsi'] + 'type', ns['xsd'] + 'string')
        facet_value_node.text = facet


def _check_shape_list(collection, indexable_element, direction, property_array, points, count):
    shape_list = collection.supporting_shape(indexable_element = indexable_element, direction = direction)
    if shape_list is not None:
        if count > 1:
            shape_list.append(count)
        if points:
            shape_list.append(3)
        if property_array is not None:
            assert tuple(shape_list) == property_array.shape, \
                f'property array shape {property_array.shape} is not the expected {tuple(shape_list)}'
